# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Expressions describe a calculation or evaluation of some sort.

It is defined as an expression tree of binary and unary operators, and functions.

Expressions are evaluated against an entire morsel at a time.
"""
from dataclasses import dataclass
from dataclasses import field
from enum import Enum
from typing import Any

import numpy
import pyarrow
from orso.cityhash import CityHash64
from pyarrow import Table

from opteryx.functions import FUNCTIONS
from opteryx.functions.binary_operators import binary_operations
from opteryx.functions.unary_operations import UNARY_OPERATIONS
from opteryx.models import Columns
from opteryx.third_party.pyarrow_ops.ops import filter_operations
from opteryx.third_party.pyarrow_ops.ops import filter_operations_for_display

# These are bit-masks
LOGICAL_TYPE: int = int("0001", 2)
INTERNAL_TYPE: int = int("0010", 2)
LITERAL_TYPE: int = int("0100", 2)


def format_expression(root):
    if root is None:
        return "null"

    node_type = root.token_type
    _map: dict = {}

    # LITERAL TYPES
    if node_type & LITERAL_TYPE == LITERAL_TYPE:
        if node_type == NodeType.LITERAL_VARCHAR:
            return "'" + root.value.replace("'", "'") + "'"
        if node_type == NodeType.LITERAL_TIMESTAMP:
            return "'" + str(root) + "'"
        if node_type == NodeType.LITERAL_INTERVAL:
            return "<INTERVAL>"
        if node_type == NodeType.LITERAL_NONE:
            return "null"
        return str(root.value)
    # INTERAL IDENTIFIERS
    if node_type & INTERNAL_TYPE == INTERNAL_TYPE:
        if node_type in (
            NodeType.FUNCTION,
            NodeType.AGGREGATOR,
            NodeType.COMPLEX_AGGREGATOR,
        ):
            if root.value == "CASE":
                con = [format_expression(a) for a in root.parameters[0].value]
                vals = [format_expression(a) for a in root.parameters[1].value]
                return "CASE " + "".join([f"WHEN {c} THEN {v} " for c, v in zip(con, vals)]) + "END"
            if root.value == "ARRAY_AGG":
                distinct = "DISTINCT " if root.parameters[1] else ""
                order = f" ORDER BY {root.parameters[2]}" if root.parameters[2] else ""
                limit = f" LIMIT {root.parameters[3]}" if root.parameters[3] else ""
                return f"{root.value.upper()}({distinct}{format_expression(root.parameters[0])}{order}{limit})"
            return (
                f"{root.value.upper()}({','.join([format_expression(e) for e in root.parameters])})"
            )
        if node_type == NodeType.WILDCARD:
            return "*"
        if node_type == NodeType.BINARY_OPERATOR:
            _map = {
                "StringConcat": "||",
                "Plus": "+",
                "Minus": "-",
                "Multiply": "*",
                "Divide": "/",
            }
            return f"{format_expression(root.left)}{_map.get(root.value, root.value)}{format_expression(root.right)}"
    if node_type == NodeType.COMPARISON_OPERATOR:
        _map = {"Eq": "=", "Lt": "<", "Gt": ">", "NotEq": "!=", "BitwiseOr": "|"}
        return f"{format_expression(root.left)}{_map.get(root.value, root.value)}{format_expression(root.right)}"
    if node_type == NodeType.UNARY_OPERATOR:
        _map = {"IsNull": "%s IS NULL", "IsNotNull": "%s IS NOT NULL"}
        return _map.get(root.value, root.value + "(%s)").replace(
            "%s", format_expression(root.centre)
        )
    if node_type == NodeType.NOT:
        return f"NOT {format_expression(root.centre)}"
    if node_type in (NodeType.AND, NodeType.OR, NodeType.XOR):
        _map = {
            NodeType.AND: "AND",
            NodeType.OR: "OR",
            NodeType.XOR: "XOR",
        }  # type:ignore
        return f"{format_expression(root.left)} {_map[node_type]} {format_expression(root.right)}"

    return str(root.value)


class NodeType(int, Enum):
    """
    The types of Nodes we will see.

    The second nibble (4 bits) is a category marker, the first nibble is just an
    enumeration of the values in that category.

    This allows us to use bitmasks to add a category to the enumerations.
    """

    # fmt:off

    # 00000000
    UNKNOWN: int = 0

    # LOGICAL OPERATORS
    # nnnn0001
    AND: int = 17
    OR: int = 33
    XOR: int = 49
    NOT: int = 65 ## 0100 0001

    # INTERAL IDENTIFIERS
    # nnnn0010
    WILDCARD: int = 18
    COMPARISON_OPERATOR: int = 34
    BINARY_OPERATOR: int = 50
    UNARY_OPERATOR: int = 66
    FUNCTION: int = 82
    IDENTIFIER: int = 98
    SUBQUERY: int = 114
    NESTED: int = 130
    AGGREGATOR:int = 146
    COMPLEX_AGGREGATOR: int = 162
    EXPRESSION_LIST:int = 178  # 1011 0010


    # LITERAL TYPES
    # nnnn0100
    LITERAL_NUMERIC: int = 20
    LITERAL_VARCHAR: int = 36
    LITERAL_BOOLEAN: int = 52
    LITERAL_INTERVAL: int = 68
    LITERAL_LIST: int = 84
    LITERAL_STRUCT: int = 100
    LITERAL_TIMESTAMP: int = 116
    LITERAL_NONE: int = 132
    LITERAL_TABLE: int = 148  # 1001 0100

    # fmt:on


NUMPY_TYPES = {
    NodeType.LITERAL_NUMERIC: numpy.dtype("float64"),
    NodeType.LITERAL_VARCHAR: numpy.unicode_(),
    NodeType.LITERAL_BOOLEAN: numpy.dtype("?"),
    NodeType.LITERAL_INTERVAL: numpy.dtype("m"),
    NodeType.LITERAL_LIST: numpy.dtype("O"),
    NodeType.LITERAL_STRUCT: numpy.dtype("O"),
    NodeType.LITERAL_TIMESTAMP: numpy.dtype("datetime64[us]"),
}


@dataclass
class ExpressionTreeNode:
    token_type: NodeType
    value: Any = None
    left: Any = None  # ExpressionTreeNode
    right: Any = None  # ExpressionTreeNode
    centre: Any = None  # ExpressionTreeNode
    parameters: list = None
    alias: list = field(default_factory=list)

    def __repr__(self):
        return (
            f"<ExpressionTreeNode {str(self.token_type).upper()}"
            f"{' `' + str(self.value) + '`' if self.value else ''} "
            f"{'L' if self.left is not None else '.'}"
            f"{'C' if self.centre is not None else '.'}"
            f"{'R' if self.right is not None else '.'}"
            f"{('[' + str(len(self.parameters)) + ']') if self.parameters is not None else '.'} "
            f"({id(self)})>"
        )

    def __str__(self):
        return str(self.value)


def _inner_evaluate(root: ExpressionTreeNode, table: Table, columns, for_display: bool = False):
    node_type = root.token_type

    # BOOLEAN OPERATORS
    if node_type & LOGICAL_TYPE == LOGICAL_TYPE:
        left, right, centre = None, None, None

        if root.left is not None:
            left = _inner_evaluate(root.left, table, columns, for_display)
        if root.right is not None:
            right = _inner_evaluate(root.right, table, columns, for_display)
        if root.centre is not None:
            centre = _inner_evaluate(root.centre, table, columns, for_display)

        if node_type == NodeType.AND:
            return pyarrow.compute.and_(left, right)
        if node_type == NodeType.OR:
            return pyarrow.compute.or_(left, right)
        if node_type == NodeType.NOT:
            null_val = pyarrow.compute.is_null(centre, nan_is_null=True)
            if any(null_val) and isinstance(centre, numpy.ndarray):
                null_val = numpy.invert(null_val)
                results_mask = centre.compress(null_val)
                results_mask = numpy.invert(results_mask)
                results = numpy.full(len(centre), -1)
                results[numpy.nonzero(null_val)] = results_mask
                return [bool(r) if r != -1 else None for r in results]
            return numpy.invert(centre)

        if node_type == NodeType.XOR:
            return pyarrow.compute.xor(left, right)

    # INTERAL IDENTIFIERS
    if node_type & INTERNAL_TYPE == INTERNAL_TYPE:
        if node_type == NodeType.FUNCTION:
            parameters = [
                _inner_evaluate(param, table, columns, for_display) for param in root.parameters
            ]
            # zero parameter functions get the number of rows as the parameter
            if len(parameters) == 0:
                parameters = [table.num_rows]
            result = FUNCTIONS[root.value](*parameters)
            if isinstance(result, list):
                result = numpy.array(result)
            return result
        if node_type in (NodeType.AGGREGATOR, NodeType.COMPLEX_AGGREGATOR):
            # detected as an aggregator, but here it's an identifier because it
            # will have already been evaluated
            node_type = NodeType.IDENTIFIER
            root.value = format_expression(root)
            root.token_type = NodeType.IDENTIFIER
        if node_type == NodeType.IDENTIFIER:
            if root.value in table.column_names:
                mapped_column = root.value
            else:
                mapped_column = columns.get_column_from_alias(root.value, only_one=True)
            return table[mapped_column].to_numpy()
        if node_type == NodeType.COMPARISON_OPERATOR:
            left = _inner_evaluate(root.left, table, columns, for_display)
            right = _inner_evaluate(root.right, table, columns, for_display)
            if for_display:
                return filter_operations_for_display(left, root.value, right)
            return filter_operations(left, root.value, right)
        if node_type == NodeType.BINARY_OPERATOR:
            left = _inner_evaluate(root.left, table, columns, for_display)
            right = _inner_evaluate(root.right, table, columns, for_display)
            return binary_operations(left, root.value, right)
        if node_type == NodeType.WILDCARD:
            numpy.full(table.num_rows, "*", dtype=numpy.unicode_)
        if node_type == NodeType.SUBQUERY:
            # we should have a query plan here
            sub = root.value.execute()
            return pyarrow.concat_tables(sub, promote=True)
        if node_type == NodeType.NESTED:
            return _inner_evaluate(root.centre, table, columns, for_display)
        if node_type == NodeType.UNARY_OPERATOR:
            centre = _inner_evaluate(root.centre, table, columns, for_display)
            return UNARY_OPERATIONS[root.value](centre)
        if node_type == NodeType.EXPRESSION_LIST:
            values = [_inner_evaluate(val, table, columns) for val in root.value]
            return values

    # LITERAL TYPES
    if node_type & LITERAL_TYPE == LITERAL_TYPE:
        # if it's a literal value, return it once for every value in the table
        if node_type == NodeType.LITERAL_LIST:
            # this isn't as fast as .full - but lists and strings are problematic
            return numpy.array([root.value] * table.num_rows)
        if node_type == NodeType.LITERAL_VARCHAR:
            return numpy.array([root.value] * table.num_rows, dtype=numpy.unicode_)
        if node_type == NodeType.LITERAL_INTERVAL:
            return pyarrow.array([root.value] * table.num_rows)
        if node_type == NodeType.LITERAL_NONE:
            return numpy.full(table.num_rows, numpy.nan)
        return numpy.full(
            shape=table.num_rows, fill_value=root.value, dtype=NUMPY_TYPES[node_type]
        )  # type:ignore


def evaluate(expression: ExpressionTreeNode, table: Table, for_display: bool = False):
    columns = Columns(table)
    result = _inner_evaluate(root=expression, table=table, columns=columns, for_display=for_display)

    if not isinstance(result, (pyarrow.Array, numpy.ndarray)):
        result = numpy.array(result)
    return result


def get_all_nodes_of_type(root, select_nodes):
    """
    Walk a expression tree collecting all the nodes of a specified type.
    """
    if not isinstance(root, list):
        root = [root]

    identifiers = []
    for node in root:
        if node.token_type in select_nodes:
            identifiers.append(node)
        if node.left:
            identifiers.extend(get_all_nodes_of_type(node.left, select_nodes))
        if node.centre:
            identifiers.extend(get_all_nodes_of_type(node.centre, select_nodes))
        if node.right:
            identifiers.extend(get_all_nodes_of_type(node.right, select_nodes))
        if node.parameters:
            for parameter in node.parameters:
                if isinstance(parameter, ExpressionTreeNode):
                    identifiers.extend(get_all_nodes_of_type(parameter, select_nodes))

    return identifiers


def evaluate_and_append(expressions, table: Table, seed: str = None):
    """
    Evaluate an expression and add it to the table.

    This needs to be able to deal with and avoid cascading problems where field names
    are duplicated, this is most common when performing many joins on the same table.
    """

    columns = Columns(table)
    return_expressions = []

    for statement in expressions:
        if statement.token_type in (
            NodeType.FUNCTION,
            NodeType.BINARY_OPERATOR,
            NodeType.COMPARISON_OPERATOR,
            NodeType.UNARY_OPERATOR,
            NodeType.NOT,
            NodeType.AND,
            NodeType.OR,
            NodeType.XOR,
        ) or (statement.token_type & LITERAL_TYPE == LITERAL_TYPE):
            new_column_name = format_expression(statement)
            raw_column_name = new_column_name

            # avoid clashes in column names
            alias = statement.alias
            if not alias:
                alias = [new_column_name]
            if seed is not None:
                alias.append(new_column_name)
                new_column_name = hex(CityHash64(seed + new_column_name))

            # if we've already been evaluated - don't do it again
            if len(columns.get_column_from_alias(raw_column_name)) > 0:
                statement = ExpressionTreeNode(
                    NodeType.IDENTIFIER, value=raw_column_name, alias=alias
                )
                return_expressions.append(statement)
                continue

            # do the evaluation
            new_column = evaluate(statement, table, True)

            # some activities give us masks rather than the values, if we don't have
            # enough values, assume it's a mask
            if len(new_column) < table.num_rows or statement.token_type in (
                NodeType.UNARY_OPERATOR,
            ):
                bool_list = numpy.full(table.num_rows, False)
                bool_list[new_column] = True
                new_column = bool_list

            # large arrays appear to have a bug in PyArrow where they're automatically
            # converted to a chunked array, but the internal function can't handle
            # chunked arrays - 50Mb columns are rare when we have 64Mb morsels.
            if new_column.nbytes > 50000000:
                new_column = [[i] for i in new_column]
            else:
                new_column = [new_column]

            table = table.append_column(new_column_name, new_column)

            # add the column to the schema and because it's been evaluated and added to
            # table, it's an INDENTIFIER rather than a FUNCTION
            columns.add_column(new_column_name)
            columns.add_alias(new_column_name, alias)
            columns.set_preferred_name(new_column_name, alias[0])

            statement = ExpressionTreeNode(NodeType.IDENTIFIER, value=new_column_name, alias=alias)

        return_expressions.append(statement)

    table = columns.apply(table)

    return columns, return_expressions, table


def deduplicate_list_of_nodes(nodes):
    seen = set()
    deduped = []
    for column in nodes:
        if column.value not in seen:
            deduped.append(column)
            seen.add(column.value)
            if column.alias:
                seen.update(column.alias)
    return deduped



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Selection Node

This is a SQL Query Execution Plan Node.
"""
import time
from typing import Iterable

import numpy
import pyarrow
from pyarrow import Table

from opteryx.exceptions import SqlError
from opteryx.managers.expression import evaluate
from opteryx.managers.expression import format_expression
from opteryx.models import QueryProperties
from opteryx.operators import BasePlanNode


class SelectionNode(BasePlanNode):
    def __init__(self, properties: QueryProperties, **config):
        super().__init__(properties=properties)
        self.filter = config.get("filter")

    @property
    def config(self):  # pragma: no cover
        return format_expression(self.filter)

    @property
    def name(self):  # pragma: no cover
        return "Selection"

    def execute(self) -> Iterable:
        if len(self._producers) != 1:  # pragma: no cover
            raise SqlError(f"{self.name} on expects a single producer")

        morsels = self._producers[0]  # type:ignore
        schema = None
        at_least_one = False

        # we should always have a filter - but no harm in checking
        if self.filter is None:
            yield from morsels.execute()
            return

        for morsel in morsels.execute():
            if schema is None:
                schema = morsel.schema

            if morsel.num_rows == 0:
                continue

            start_selection = time.time_ns()
            mask = evaluate(self.filter, morsel, False)
            self.statistics.time_evaluating += time.time_ns() - start_selection

            # if the mask is a boolean array, we've called a function that
            # returns booleans
            if isinstance(mask, pyarrow.lib.BooleanArray) or (
                isinstance(mask, numpy.ndarray) and mask.dtype == numpy.bool_
            ):
                mask = numpy.nonzero(mask)[0]

            self.statistics.time_selecting += time.time_ns() - start_selection

            # if there's no matching rows, just drop the morsel
            if mask.size > 0:
                yield morsel.take(pyarrow.array(mask))
                at_least_one = True

        # we need to send something to the next operator, send an empty table
        if not at_least_one:
            yield pyarrow.Table.from_arrays([[] for i in schema.names], schema=schema)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

import numpy
from pyarrow import compute

from opteryx.exceptions import ProgrammingError


def string_slicer_left(arr, length):
    """
    Slice a list of strings from the left
    """
    if len(arr) == 0:
        return [[]]
    if not hasattr(length, "__iter__"):
        length = [length] * len(arr)
    if hasattr(arr, "to_numpy"):
        arr = arr.to_numpy(False)
    return [None if s is None else s[: int(length[i])] for i, s in enumerate(arr)]


def string_slicer_right(arr, length):
    """
    Slice a list of strings from the right
    """
    if len(arr) == 0:
        return [[]]
    if not hasattr(length, "__iter__"):
        length = [length] * len(arr)
    if hasattr(arr, "to_numpy"):
        arr = arr.to_numpy(False)
    return [None if s is None else s[-int(length[i]) :] for i, s in enumerate(arr)]


def split(arr, delimiter=",", limit=None):
    """
    Slice a list of strings from the right
    """
    if not isinstance(delimiter, str):
        delimiter = delimiter[0]
    if limit is not None:
        limit = int(limit[0]) - 1
        if limit < 0:
            raise ProgrammingError("`SPLIT` limit parameter must be greater than zero.")
    return compute.split_pattern(arr, pattern=delimiter, max_splits=limit)


def soundex(arr):
    from opteryx.third_party.fuzzy.soundex import Soundex

    _soundex = Soundex(4)
    interim = ["0000"] * arr.size

    for index, string in enumerate(arr):
        if string:
            interim[index] = _soundex(string)
        else:
            interim[index] = None

    return numpy.array(interim, dtype=numpy.str_)


def get_md5(item):
    """calculate MD5 hash of a value"""
    import hashlib  # delay the import - it's rarely needed

    if item is None:
        return None

    return hashlib.md5(str(item).encode()).hexdigest()  # nosec - meant to be MD5


def get_sha1(item):
    """calculate SHA1 hash of a value"""
    import hashlib  # delay the import - it's rarely needed

    if item is None:
        return None

    return hashlib.sha1(str(item).encode()).hexdigest()


def get_sha224(item):
    """calculate SHA256 hash of a value"""
    import hashlib  # delay the import - it's rarely needed

    if item is None:
        return None

    return hashlib.sha224(str(item).encode()).hexdigest()


def get_sha256(item):
    """calculate SHA256 hash of a value"""
    import hashlib  # delay the import - it's rarely needed

    if item is None:
        return None

    return hashlib.sha256(str(item).encode()).hexdigest()


def get_sha384(item):
    """calculate SHA256 hash of a value"""
    import hashlib  # delay the import - it's rarely needed

    if item is None:
        return None

    return hashlib.sha384(str(item).encode()).hexdigest()


def get_sha512(item):
    """calculate SHA512 hash of a value"""
    import hashlib  # delay the import - it's rarely needed

    if item is None:
        return None

    return hashlib.sha512(str(item).encode()).hexdigest()


def get_base64_encode(item):
    """calculate BASE64 encoding of a string"""
    import base64

    if item is None:
        return None

    if not isinstance(item, bytes):
        item = str(item).encode()
    return base64.b64encode(item).decode("UTF8")


def get_base64_decode(item):
    """calculate BASE64 encoding of a string"""
    import base64

    if item is None:
        return None

    if not isinstance(item, bytes):
        item = str(item).encode()
    return base64.b64decode(item).decode("UTF8")


def get_base85_encode(item):
    """calculate BASE85 encoding of a string"""
    import base64

    if item is None:
        return None

    if not isinstance(item, bytes):
        item = str(item).encode()
    return base64.b85encode(item).decode("UTF8")


def get_base85_decode(item):
    """calculate BASE85 encoding of a string"""
    import base64

    if item is None:
        return None

    if not isinstance(item, bytes):
        item = str(item).encode()
    return base64.b85decode(item).decode("UTF8")


def get_hex_encode(item):
    """calculate HEX encoding of a string"""
    import base64

    if item is None:
        return None

    if not isinstance(item, bytes):
        item = str(item).encode()
    return base64.b16encode(item).decode("UTF8")


def get_hex_decode(item):
    """calculate HEX encoding of a string"""
    import base64

    if item is None:
        return None

    if not isinstance(item, bytes):
        item = str(item).encode()
    return base64.b16decode(item).decode("UTF8")


def concat(list_values):
    """concatenate a list of strings"""
    result: List = []
    for row in list_values:
        if row is None:
            result.append(None)
        else:
            row = row.astype(dtype=numpy.str_)
            result.append("".join(row))
    return result


def concat_ws(separator, list_values):
    """concatenate a list of strings with a separator"""
    result: List = []
    if len(separator) > 0:
        separator = separator[0]
        if separator is None:
            return None
    for row in list_values:
        if row is None:
            result.append(None)
        else:
            row = row.astype(dtype=numpy.str_)
            result.append(separator.join(row))
    return result


def starts_w(arr, test):
    return compute.starts_with(arr, test[0])


def ends_w(arr, test):
    return compute.ends_with(arr, test[0])


def substring(arr, from_pos, count):
    if len(arr) == 0:
        return [[]]

    def _inner(val, _from, _for):
        _from = int(_from)
        if _from > 0:
            _from -= 1
        _for = int(_for) if _for == _for else None  # nosec
        if _for is None:
            return val[_from:]
        return val[_from : _for + _from]

    def _outer():
        for index, value in enumerate(arr):
            yield _inner(value, from_pos[index], count[index])

    return list(_outer())


def position(sub, string):
    """
    Returns the starting position of the first instance of substring in string. Positions start with 1. If not found, 0 is returned.
    """
    return string.find(sub) + 1


def trim(*args):
    if len(args) == 1:
        return compute.utf8_trim_whitespace(args[0])
    return compute.utf8_trim(args[0], args[1][0])


def ltrim(*args):
    if len(args) == 1:
        return compute.utf8_ltrim_whitespace(args[0])
    return compute.utf8_ltrim(args[0], args[1][0])


def rtrim(*args):
    if len(args) == 1:
        return compute.utf8_rtrim_whitespace(args[0])
    return compute.utf8_rtrim(args[0], args[1][0])


def levenshtein(a, b):
    from opteryx.third_party.levenshtein.levenshtein import levenshtein as lev

    def _outer():
        for index, value in enumerate(a):
            yield lev(value, b[index])

    return list(_outer())



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
MinIo Reader - also works with AWS
"""
import io
import os

from opteryx.connectors import BaseBlobStorageAdapter
from opteryx.exceptions import MissingDependencyError
from opteryx.exceptions import UnmetRequirementError
from opteryx.utils import paths


class AwsS3Connector(BaseBlobStorageAdapter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        try:
            from minio import Minio  # type:ignore
        except ImportError as err:  # pragma: no cover
            raise MissingDependencyError(err.name) from err

        super().__init__(**kwargs)

        end_point = os.environ.get("MINIO_END_POINT")
        access_key = os.environ.get("MINIO_ACCESS_KEY")
        secret_key = os.environ.get("MINIO_SECRET_KEY")
        secure = str(os.environ.get("MINIO_SECURE", "TRUE")).lower() == "true"

        if end_point is None:  # pragma: no cover
            raise UnmetRequirementError(
                "MinIo (S3) adapter requires MINIO_END_POINT, MINIO_ACCESS_KEY and MINIO_SECRET_KEY set in environment variables."
            )

        self.minio = Minio(end_point, access_key, secret_key, secure=secure)

    def get_blob_list(self, partition):
        bucket, object_path, _, _ = paths.get_parts(partition)
        blobs = self.minio.list_objects(bucket_name=bucket, prefix=object_path, recursive=True)
        return [
            bucket + "/" + blob.object_name for blob in blobs if not blob.object_name.endswith("/")
        ]

    def read_blob(self, blob_name):
        try:
            bucket, object_path, name, extension = paths.get_parts(blob_name)
            stream = self.minio.get_object(bucket, object_path + name + extension)
            return io.BytesIO(stream.read())
        finally:
            stream.close()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Date Utilities
"""
import re
from datetime import date
from datetime import datetime
from datetime import time as dtime
from datetime import timedelta
from datetime import timezone
from functools import lru_cache
from typing import Union

import numpy

TIMEDELTA_REGEX = (
    r"((?P<years>\d+)\s?(?:ys?|yrs?|years?))?\s*"
    r"((?P<months>\d+)\s?(?:mo|mons?|mths?|months?))?\s*"
    r"((?P<weeks>\d+)\s?(?:w|wks?|weeks?))?\s*"
    r"((?P<days>\d+)\s?(?:d|days?))?\s*"
    r"((?P<hours>\d+)\s?(?:h|hrs?|hours?))?\s*"
    r"((?P<minutes>\d+)\s?(?:m|mins?|minutes?))?\s*"
    r"((?P<seconds>\d+)\s?(?:s|secs?|seconds?))?\s*"
)

TIMEDELTA_PATTERN = re.compile(TIMEDELTA_REGEX, re.IGNORECASE)


def add_months(dte, num_months):
    new_year, new_month = divmod(dte.month - 1 + num_months, 12)
    new_year += dte.year
    new_month += 1
    last_day_of_month = (datetime(new_year, new_month % 12 + 1, 1) - timedelta(days=1)).day
    new_day = min(dte.day, last_day_of_month)
    return datetime(new_year, new_month, new_day)


def add_interval(current_date: Union[date, datetime], interval: str) -> Union[date, datetime]:
    """
    Parses a human readable timedelta (3d5h19m) into a datetime.timedelta.
    """
    match = TIMEDELTA_PATTERN.match(interval)
    if match:
        parts = {k: int(v) for k, v in match.groupdict().items() if v}
        # time delta doesn't include weeks, months or years
        if "weeks" in parts:
            weeks = parts.pop("weeks")
            current_date = current_date + timedelta(days=weeks * 7)
        if "months" in parts:
            months = parts.pop("months")
            current_date = add_months(current_date, months)
        if "years" in parts:
            # need to avoid 29th Feb problems, so can't just say year - year
            years = parts.pop("years")
            current_date = add_months(current_date, 12 * years)
        if parts:
            return current_date + timedelta(**parts)
        return current_date
    raise ValueError(f"Unable to interpret interval - {interval}")


def date_range(start, end, interval: str):
    """ """
    start = parse_iso(start)
    end = parse_iso(end)

    if start is end or start == end or start > end:
        raise ValueError("Cannot create an series with the provided start and end dates")

    cursor = start
    while cursor <= end:
        yield cursor
        cursor = add_interval(cursor, interval)


@lru_cache(128)
def parse_iso(value):
    date_separators = ("-", ":")
    # date validation at speed is hard, dateutil is great but really slow, this is fast
    # but error-prone. It assumes it is a date or it really nothing like a date.
    # Making that assumption - and accepting the consequences - we can convert upto
    # three times faster than dateutil.
    #
    # valid formats (not exhaustive):
    #
    #   YYYY-MM-DD                 <- date
    #   YYYY-MM-DD HH:MM           <- date and time, no seconds
    #   YYYY-MM-DDTHH:MM           <- date and time, T separator
    #   YYYY-MM-DD HH:MM:SS        <- date and time with seconds
    #   YYYY-MM-DD HH:MM:SS.mmmm   <- date and time with milliseconds
    #
    # If the last character is a Z, we ignore it.
    try:
        input_type = type(value)

        if input_type in (int, numpy.int64):
            value = numpy.datetime64(int(value), "s").astype(datetime)
            input_type = type(value)

        if input_type == numpy.datetime64:
            # this can create dates rather than datetimes, so don't return yet
            value = value.astype(datetime)
            input_type = type(value)

        if input_type == datetime:
            return value
        if input_type == date:
            return datetime.combine(value, dtime.min)
        if input_type in (int, float):
            return datetime.fromtimestamp(value)
        if input_type == str and 10 <= len(value) <= 33:
            if value[-1] == "Z":
                value = value[:-1]
            if "+" in value:
                value = value.split("+")[0]
                if not 10 <= len(value) <= 28:
                    return None
            val_len = len(value)
            if not value[4] in date_separators or not value[7] in date_separators:
                return None
            if val_len == 10:
                # YYYY-MM-DD
                return datetime(
                    *map(int, [value[:4], value[5:7], value[8:10]])  # type:ignore
                )
            if val_len >= 16:
                if not (value[10] in ("T", " ") and value[13] in date_separators):
                    return False
                if val_len >= 19 and value[16] in date_separators:
                    # YYYY-MM-DD HH:MM:SS
                    return datetime(
                        *map(  # type:ignore
                            int,
                            [
                                value[:4],  # YYYY
                                value[5:7],  # MM
                                value[8:10],  # DD
                                value[11:13],  # HH
                                value[14:16],  # MM
                                value[17:19],  # SS
                            ],
                        )
                    )
                if val_len == 16:
                    # YYYY-MM-DD HH:MM
                    return datetime(
                        *map(  # type:ignore
                            int,
                            [
                                value[:4],
                                value[5:7],
                                value[8:10],
                                value[11:13],
                                value[14:16],
                            ],
                        )
                    )
        return None
    except (ValueError, TypeError):
        return None


EPOCH: date = datetime(1970, 1, 1, tzinfo=timezone.utc)


def date_trunc(truncate_to, dt):
    # convert acceptable non datetime values to datetime
    dt = parse_iso(dt)

    if not isinstance(truncate_to, str):
        truncate_to = truncate_to[0]  # [#325]

    # [#711]
    truncate_to = str(truncate_to).lower()

    if truncate_to == "year":
        return datetime(dt.year, 1, 1, tzinfo=dt.tzinfo)
    elif truncate_to == "quarter":
        quarter = (dt.month - 1) // 3 + 1
        return datetime(dt.year, 3 * quarter - 2, 1, tzinfo=dt.tzinfo)
    elif truncate_to == "month":
        return datetime(dt.year, dt.month, 1, tzinfo=dt.tzinfo)
    elif truncate_to == "week":
        days_since_monday = dt.weekday()
        monday = dt - timedelta(days=days_since_monday)
        return date_trunc("day", monday)
    elif truncate_to == "day":
        return datetime(dt.year, dt.month, dt.day, tzinfo=dt.tzinfo)
    elif truncate_to == "hour":
        return datetime(dt.year, dt.month, dt.day, dt.hour, tzinfo=dt.tzinfo)
    elif truncate_to == "minute":
        return datetime(dt.year, dt.month, dt.day, dt.hour, dt.minute, tzinfo=dt.tzinfo)
    elif truncate_to == "second":
        return datetime(dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second, tzinfo=dt.tzinfo)
    else:
        raise ValueError("Invalid unit: {}".format(truncate_to))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import orjson

from opteryx import operators
from opteryx.components.logical_planner import builders
from opteryx.components.logical_planner import custom_builders
from opteryx.connectors import DiskConnector
from opteryx.connectors import connector_factory
from opteryx.exceptions import DatabaseError
from opteryx.exceptions import ProgrammingError
from opteryx.exceptions import SqlError
from opteryx.managers.expression import ExpressionTreeNode
from opteryx.managers.expression import NodeType
from opteryx.managers.expression import deduplicate_list_of_nodes
from opteryx.managers.expression import get_all_nodes_of_type
from opteryx.models import ExecutionTree
from opteryx.utils import paths


def explain_query(ast, properties):
    # we're handling two plans here:
    # - plan - this is the plan for the query we're exlaining
    # - my_plan - this is the plan for this query

    from opteryx.components.query_planner import QueryPlanner

    query_planner = QueryPlanner(properties=properties)
    plan = query_planner.create_logical_plan(ast["Explain"]["statement"])
    plan = query_planner.optimize_plan(plan)
    my_plan = ExecutionTree()
    explain_node = operators.ExplainNode(properties, query_plan=plan)
    my_plan.add_node("explain", explain_node)
    return my_plan


def select_query(ast, properties):
    """
    The planner creates the naive query plan.

    The goal here is to create a plan that's only guarantee is the response is correct.
    It doesn't try to make it performant, low-memory or any other measure of 'good'
    beyond correctness.
    """
    plan = ExecutionTree()

    properties.ctes = custom_builders.extract_ctes(ast["Query"], properties.qid)

    all_identifiers = (
        set(custom_builders.extract_identifiers(ast)) - custom_builders.WELL_KNOWN_HINTS
    )
    try:
        _relations = list(
            custom_builders.extract_relations(
                ast["Query"]["body"]["Select"]["from"], properties.qid
            )
        )
    except IndexError:
        _relations = []

    # if we have no relations, use the $no_table relation
    if len(_relations) == 0:
        _relations = [
            custom_builders.RelationDescription(
                dataset="$no_table", kind="Internal", cache=properties.cache
            )
        ]

    # We always have a data source - even if it's 'no table'
    relation = _relations[0]

    reader = None
    if isinstance(relation.dataset, str) and relation.dataset in properties.ctes:
        # CTEs look like subqueries
        relation.kind = "SubQuery"
        relation.alias = relation.dataset
        relation.dataset = properties.ctes[relation.dataset]
    if relation.kind == "File":
        reader = DiskConnector(prefix="")
    elif relation.kind == "External":
        # external comes in different flavours
        reader = connector_factory(relation.dataset)
        relation.kind = reader.__mode__

    plan.add_node(
        "from",
        operators.reader_factory(relation.kind)(
            properties=properties,
            alias=relation.alias,
            dataset=relation.dataset,
            reader=reader,
            cache=relation.cache,
            start_date=relation.start_date,
            end_date=relation.end_date,
            hints=relation.hints,
            selection=all_identifiers,
        ),
    )
    last_node = "from"

    _joins = list(custom_builders.extract_joins(ast, properties.qid))
    if len(_joins) == 0 and len(_relations) == 2:
        # If there's no explicit JOIN but the query has two relations, we
        # use a CROSS JOIN
        _joins = [("CrossJoin", _relations[1], None, None)]
    for join_id, _join in enumerate(_joins):
        if _join:
            join_type, right, join_on, join_using = _join
            if join_type == "CrossJoin" and right.kind == "Function":
                join_type = "CrossJoinUnnest"
            else:
                dataset = right.dataset
                if isinstance(dataset, ExecutionTree):
                    mode = "Blob"  # subqueries are here due to legacy reasons
                    reader = None
                elif isinstance(dataset, dict) and dataset.get("function") is not None:
                    mode = "Function"
                    reader = None
                elif dataset[0:1] == "$":
                    mode = "Internal"
                    reader = None
                elif dataset in properties.ctes:
                    # CTEs look like subqueries
                    mode = "SubQuery"  # subqueries are here due to legacy reasons
                    reader = None
                    right.alias = right.dataset
                    right.dataset = properties.ctes[dataset]
                elif paths.is_file(dataset):
                    mode = "File"
                else:
                    reader = connector_factory(dataset)
                    mode = reader.__mode__

                # Otherwise, the right table needs to come from the Reader
                right = operators.reader_factory(mode)(
                    properties=properties,
                    dataset=right.dataset,
                    alias=right.alias,
                    reader=reader,
                    cache=relation.cache,
                    start_date=right.start_date,
                    end_date=right.end_date,
                    hints=right.hints,
                )

            join_node = operators.join_factory(join_type)
            if join_node is None:
                raise SqlError(f"Join type not supported - `{_join[0]}`")

            plan.add_node(
                f"join-{join_id}",
                join_node(
                    properties=properties,
                    join_type=join_type,
                    join_on=join_on,
                    join_using=join_using,
                ),
            )
            plan.add_edge(last_node, f"join-{join_id}")

            plan.add_node(f"join-{join_id}-right", right)
            plan.add_edge(f"join-{join_id}-right", f"join-{join_id}", "right")

            last_node = f"join-{join_id}"

    _selection = builders.build(ast["Query"]["body"]["Select"]["selection"])
    if _selection:
        plan.add_node(
            "where",
            operators.SelectionNode(properties, filter=_selection),
        )
        plan.add_edge(last_node, "where")
        last_node = "where"

    _projection = builders.build(ast["Query"]["body"]["Select"]["projection"])
    _groups = builders.build(ast["Query"]["body"]["Select"]["group_by"])
    if _groups or get_all_nodes_of_type(
        _projection, select_nodes=(NodeType.AGGREGATOR, NodeType.COMPLEX_AGGREGATOR)
    ):
        _aggregates = _projection.copy()
        if isinstance(_aggregates, dict):
            raise SqlError("GROUP BY cannot be used with SELECT *")
        plan.add_node(
            "agg",
            operators.AggregateNode(properties, aggregates=_aggregates, groups=_groups),
        )
        plan.add_edge(last_node, "agg")
        last_node = "agg"

    _having = builders.build(ast["Query"]["body"]["Select"]["having"])
    if _having:
        plan.add_node(
            "having",
            operators.SelectionNode(properties, filter=_having),
        )
        plan.add_edge(last_node, "having")
        last_node = "having"

    # collect ORDER BY now, so we can keep any columns in the ORDER BY clause too
    _order = custom_builders.extract_order(ast)
    reproject = None

    if _order and (_projection[0].token_type != NodeType.WILDCARD):
        order_fields = [f[0][0] for f in _order if f[0][0].token_type == NodeType.IDENTIFIER]
        reproject = _projection.copy()
        _projection.extend(order_fields)
        # aliases appear in the list as different fields here, so dedupe and see if the
        # lists are different lengths
        _projection = deduplicate_list_of_nodes(_projection)
        if len(_projection) == len(reproject):
            reproject = None

    # qualified wildcards have the qualifer in the value
    # e.g. SELECT table.* -> node.value = table
    if (_projection[0].token_type != NodeType.WILDCARD) or (_projection[0].value is not None):
        plan.add_node(
            "select",
            operators.ProjectionNode(properties, projection=_projection),
        )
        plan.add_edge(last_node, "select")
        last_node = "select"

    _distinct = custom_builders.extract_distinct(ast)
    if _distinct:
        plan.add_node("distinct", operators.DistinctNode(properties))
        plan.add_edge(last_node, "distinct")
        last_node = "distinct"

    if _order:
        plan.add_node("order", operators.SortNode(properties, order=_order))
        plan.add_edge(last_node, "order")
        last_node = "order"

    # if we need to project after the order by
    if reproject:
        plan.add_node(
            "post_order_select",
            operators.ProjectionNode(properties, projection=reproject),
        )
        plan.add_edge(last_node, "post_order_select")
        last_node = "post_order_select"

    _offset = custom_builders.extract_offset(ast)
    if _offset:
        plan.add_node(
            "offset",
            operators.OffsetNode(properties, offset=_offset),
        )
        plan.add_edge(last_node, "offset")
        last_node = "offset"

    _limit = custom_builders.extract_limit(ast)
    # 0 limit is valid
    if _limit is not None:
        plan.add_node("limit", operators.LimitNode(properties, limit=_limit))
        plan.add_edge(last_node, "limit")
        last_node = "limit"

    _insert = custom_builders.extract_into(ast)

    return plan


def set_variable_query(ast, properties):
    """put variables defined in SET statements into context"""
    key = ast["SetVariable"]["variable"][0]["value"]
    value = builders.build(ast["SetVariable"]["value"][0]["Value"])
    if key[0] == "@":  # pragma: no cover
        properties.variables[key] = value
    else:
        key = key.lower()
        if key in properties.read_only_properties:
            raise ProgrammingError(f"Invalid parameter '{key}'")
        if hasattr(properties, key):
            setattr(properties, key, value.value)
        else:
            raise ProgrammingError(
                f"Unknown parameter, variables must be prefixed with a '@' - '{key}'"
            )

    # return a plan, because it's expected
    plan = ExecutionTree()
    operator = operators.ShowValueNode(key="result", value="Complete", properties=properties)
    plan.add_node("show", operator)
    return plan


def show_columns_query(ast, properties):
    plan = ExecutionTree()
    dataset = ".".join([part["value"] for part in ast["ShowColumns"]["table_name"]])

    if dataset[0:1] == "$":
        mode = "Internal"
        reader = None
    else:
        reader = connector_factory(dataset)
        mode = reader.__mode__

    plan.add_node(
        "reader",
        operators.reader_factory(mode)(
            properties=properties,
            dataset=dataset,
            alias=None,
            reader=reader,
            cache=None,  # never read from cache
            start_date=ast["ShowColumns"]["table_name"][0]["start_date"],
            end_date=ast["ShowColumns"]["table_name"][0]["end_date"],
        ),
    )
    last_node = "reader"

    filters = custom_builders.extract_show_filter(ast["ShowColumns"])
    if filters:
        plan.add_node(
            "filter",
            operators.ColumnFilterNode(properties=properties, filter=filters),
        )
        plan.add_edge(last_node, "filter")
        last_node = "filter"

    plan.add_node(
        "columns",
        operators.ShowColumnsNode(
            properties=properties,
            full=ast["ShowColumns"]["full"],
            extended=ast["ShowColumns"]["extended"],
        ),
    )
    plan.add_edge(last_node, "columns")
    last_node = "columns"

    return plan


def show_create_query(ast, properties):
    plan = ExecutionTree()

    if ast["ShowCreate"]["obj_type"] != "Table":
        raise SqlError("SHOW CREATE only supports tables")

    dataset = ".".join([part["value"] for part in ast["ShowCreate"]["obj_name"]])

    if dataset[0:1] == "$":
        mode = "Internal"
        reader = None
    else:
        reader = connector_factory(dataset)
        mode = reader.__mode__

    plan.add_node(
        "reader",
        operators.reader_factory(mode)(
            properties=properties,
            dataset=dataset,
            alias=None,
            reader=reader,
            cache=None,  # never read from cache
            start_date=ast["ShowCreate"]["start_date"],
            end_date=ast["ShowCreate"]["end_date"],
        ),
    )
    last_node = "reader"

    plan.add_node(
        "show_create",
        operators.ShowCreateNode(properties=properties, table=dataset),
    )
    plan.add_edge(last_node, "show_create")
    last_node = "show_create"

    return plan


def show_variable_query(ast, properties):
    """
    This is the generic SHOW <variable> handler - there are specific handlers
    for some keywords after SHOW, like SHOW COLUMNS.

    SHOW <variable> only really has a single node.

    All of the keywords should up as a 'values' list in the variable in the ast.
    """

    plan = ExecutionTree()
    name_column = None

    keywords = [value["value"].upper() for value in ast["ShowVariable"]["variable"]]
    if keywords[0] == "PARAMETER":
        if len(keywords) != 2:
            raise SqlError("`SHOW PARAMETER` expects a single parameter name.")
        key = keywords[1].lower()
        if not hasattr(properties, key) or key == "variables":
            raise SqlError(f"Unknown parameter '{key}'.")
        value = getattr(properties, key)

        show_node = "show_parameter"
        node = operators.ShowValueNode(properties=properties, key=key, value=value)
        plan.add_node(show_node, node)
    elif keywords[0] == "DATABASES":
        if len(keywords) != 1:
            raise SqlError(f"`SHOW DATABASES` end expected, got '{keywords[1]}'")
        show_node = "show_databases"
        node = operators.ShowDatabasesNode(properties=properties)  # type:ignore
        plan.add_node(show_node, node)
        name_column = ExpressionTreeNode(NodeType.IDENTIFIER, value="Database")
    else:  # pragma: no cover
        raise SqlError(f"SHOW statement type not supported for `{keywords[0]}`.")

    if name_column is None:
        name_column = ExpressionTreeNode(NodeType.IDENTIFIER, value="name")

    order_by_node = operators.SortNode(
        properties=properties,
        order=[([name_column], "ascending")],
    )
    plan.add_node("order", order_by_node)
    plan.add_edge(show_node, "order")

    return plan


def show_functions_query(ast, properties):
    """show the supported functions, optionally filter them"""
    plan = ExecutionTree()

    show = operators.ShowFunctionsNode(properties=properties)
    plan.add_node("show", show)
    last_node = "show"

    filters = custom_builders.extract_show_filter(ast["ShowFunctions"])
    if filters:
        plan.add_node(
            "filter",
            operators.SelectionNode(properties=properties, filter=filters),
        )
        plan.add_edge(last_node, "filter")

    return plan


def show_variables_query(ast, properties):
    """show the known variables, optionally filter them"""
    plan = ExecutionTree()

    show = operators.ShowVariablesNode(properties=properties)
    plan.add_node("show", show)
    last_node = "show"

    filters = custom_builders.extract_show_filter(ast["ShowVariables"], "variable_name")
    if filters:
        plan.add_node(
            "filter",
            operators.SelectionNode(properties=properties, filter=filters),
        )
        plan.add_edge(last_node, "filter")

    return plan


def analyze_query(ast, properties):
    """build statistics for a table"""

    # TODO: [TARCHIA] - get a list of all of the blobs for this dataset and trigger add
    # requests to Tarchia - this will add/update the statistics for the blob

    plan = ExecutionTree()
    dataset = ".".join([part["value"] for part in ast["Analyze"]["table_name"]])

    if dataset[0:1] == "$":
        mode = "Internal"
        reader = None
    else:
        if paths.is_file(dataset):
            mode = "File"
            reader = DiskConnector(prefix="")
        else:
            reader = connector_factory(dataset)
            mode = reader.__mode__

    plan.add_node(
        "reader",
        operators.reader_factory(mode)(
            properties=properties,
            dataset=dataset,
            alias=None,
            reader=reader,
            cache=None,  # never read from cache
            start_date=ast["Analyze"]["table_name"][0]["start_date"],
            end_date=ast["Analyze"]["table_name"][0]["end_date"],
        ),
    )
    last_node = "reader"

    plan.add_node(
        "buildstats",
        operators.BuildStatisticsNode(properties=properties),
    )
    plan.add_edge(last_node, "buildstats")

    return plan


def execute_query(ast, properties):
    """execute a prepared statement"""
    try:
        statement_name = ast["Execute"]["name"]["value"]
        parameters = [builders.build(p["Value"]) for p in ast["Execute"]["parameters"]]
        prepared_statatements = orjson.loads(open("prepared_statements.json").read())
        if statement_name not in prepared_statatements:
            raise SqlError("Unable to EXECUTE prepared statement, '{statement_name}' not defined.")
        prepared_statement = prepared_statatements[statement_name]

        from opteryx.components.query_planner import QueryPlanner
        from opteryx.components.sql_rewriter.sql_rewriter import clean_statement
        from opteryx.components.sql_rewriter.sql_rewriter import remove_comments
        from opteryx.components.sql_rewriter.temporal_extraction import extract_temporal_filters

        # these would have been applied to the EXECUTE statement, we want to do them on the
        # prepared statement
        statement = remove_comments(prepared_statement)
        statement = clean_statement(statement)
        statement, properties.temporal_filters = extract_temporal_filters(statement)

        # we need to plan the prepared statement
        query_planner = QueryPlanner(properties=properties, statement=statement)
        asts = list(query_planner.parse_and_lex())
        if len(asts) != 1:
            raise SqlError("Cannot execute prepared batch statements.")
        query_planner.test_paramcount(asts, parameters)
        prepared_ast = query_planner.bind_ast(asts[0], parameters=parameters)
        plan = query_planner.create_logical_plan(prepared_ast)
        # optimize is the next step, so don't need to do it here
        return plan

    except OSError as err:
        raise DatabaseError(
            "Unable to EXECUTE prepared statement, cannot find definitions."
        ) from err


def use_query(ast, properties):
    # TODO - set the default database in the connection
    plan = ExecutionTree()
    plan.add_node(
        "no_op",
        operators.NoOpNode(properties=properties),
    )
    return plan


# wrappers for the query builders
QUERY_BUILDER = {
    "Analyze": analyze_query,
    "Execute": execute_query,
    "Explain": explain_query,
    "Query": select_query,
    "SetVariable": set_variable_query,
    "ShowColumns": show_columns_query,
    "ShowCreate": show_create_query,
    "ShowFunctions": show_functions_query,
    "ShowVariable": show_variable_query,  # generic SHOW handler
    "ShowVariables": show_variables_query,
    "Use": use_query,
}



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Decode files from a raw binary format to a PyArrow Table.
"""
from enum import Enum
from typing import List

import numpy
import pyarrow


class ExtentionType(str, Enum):
    """labels for the file extentions"""

    DATA = "DATA"
    CONTROL = "CONTROL"


def do_nothing(stream, projection=None):  # pragma: no-cover
    return stream


def _filter(filter, table):
    # notes:
    #   at this point we've not renamed any columns
    from opteryx.managers.expression import evaluate

    mask = evaluate(filter, table, False)
    return table.filter(mask)


def zstd_decoder(stream, projection: List = None, selection=None):
    """
    Read zstandard compressed JSONL files
    """
    import zstandard

    with zstandard.open(stream, "rb") as file:
        return jsonl_decoder(file, projection=projection, selection=selection)


def parquet_decoder(stream, projection: List = None, selection=None):
    """
    Read parquet formatted files
    """
    from pyarrow import parquet

    from opteryx.connectors.capabilities import predicate_pushable

    # parquet uses DNF filters
    _select = None
    if selection is not None:
        _select = predicate_pushable.to_dnf(selection)

    selected_columns = None
    if isinstance(projection, (list, set)) and "*" not in projection:
        # if we have a pushed down projection, get the list of columns from the file
        # and then only set the reader to read those
        parquet_file = parquet.ParquetFile(stream)
        # .schema_arrow is probably slower than .schema but there are instances of
        # .schema being incomplete #468
        parquet_metadata = parquet_file.schema_arrow

        if projection == {"count_*"}:
            return pyarrow.Table.from_pydict(
                {"_": numpy.full(parquet_file.metadata.num_rows, True, dtype=numpy.bool_)}
            )

        selected_columns = list(set(parquet_metadata.names).intersection(projection))
        # if nothing matched, there's been a problem - maybe HINTS confused for columns
        if len(selected_columns) == 0:  # pragma: no-cover
            selected_columns = None
    # don't prebuffer - we're already buffered as an IO Stream
    return parquet.read_table(stream, columns=selected_columns, pre_buffer=False, filters=_select)


def orc_decoder(stream, projection: List = None, selection=None):
    """
    Read orc formatted files
    """
    import pyarrow.orc as orc

    orc_file = orc.ORCFile(stream)

    selected_columns = None
    if isinstance(projection, (list, set)) and "*" not in projection:
        orc_metadata = orc_file.schema
        selected_columns = list(set(orc_metadata.names).intersection(projection))
        # if nothing matched, there's been a problem - maybe HINTS confused for columns
        if len(selected_columns) == 0:  # pragma: no-cover
            selected_columns = None

    table = orc_file.read(columns=selected_columns)
    if selection is not None:
        table = _filter(selection, table)
    return table


def jsonl_decoder(stream, projection: List = None, selection=None):
    import pyarrow.json

    table = pyarrow.json.read_json(stream)

    # the read doesn't support projection, so do it now
    if projection and "*" not in projection:
        selected_columns = list(set(table.column_names).intersection(projection))
        # if nothing matched, don't do a thing
        if len(selected_columns) > 0:
            table = table.select(selected_columns)

    if selection is not None:
        table = _filter(selection, table)
    return table


def csv_decoder(stream, projection: List = None, selection=None, delimiter=","):
    import pyarrow.csv
    from pyarrow.csv import ParseOptions

    parse_options = ParseOptions(delimiter=delimiter, newlines_in_values=True)
    table = pyarrow.csv.read_csv(stream, parse_options=parse_options)

    # the read doesn't support projection, so do it now
    if projection and "*" not in projection:
        selected_columns = list(set(table.column_names).intersection(projection))
        # if nothing matched, don't do a thing
        if len(selected_columns) > 0:
            table = table.select(selected_columns)

    if selection is not None:
        table = _filter(selection, table)
    return table


def tsv_decoder(stream, projection: List = None, selection=None):
    return csv_decoder(stream=stream, projection=projection, selection=selection, delimiter="\t")


def arrow_decoder(stream, projection: List = None, selection=None):
    import pyarrow.feather as pf

    table = pf.read_table(stream)

    # we can't get the schema before reading the file, so do selection now
    if projection and "*" not in projection:
        selected_columns = list(set(table.column_names).intersection(projection))
        # if nothing matched, don't do a thing
        if len(selected_columns) > 0:
            table = table.select(selected_columns)

    if selection is not None:
        table = _filter(selection, table)
    return table


def avro_decoder(stream, projection: List = None, selection=None):
    try:
        from avro.datafile import DataFileReader
        from avro.io import DatumReader
    except ImportError:  # pragma: no-cover
        raise Exception("`avro` is missing, please install or include in your `requirements.txt`.")

    reader = DataFileReader(stream, DatumReader())
    table = pyarrow.Table.from_pylist(list(reader))
    if projection and "*" not in projection:
        selected_columns = list(set(table.column_names).intersection(projection))
        if len(selected_columns) > 0:
            table = table.select(selected_columns)
    if selection is not None:
        table = _filter(selection, table)
    return table


KNOWN_EXTENSIONS = {
    "avro": (avro_decoder, ExtentionType.DATA),
    "complete": (do_nothing, ExtentionType.CONTROL),
    "ignore": (do_nothing, ExtentionType.CONTROL),
    "arrow": (arrow_decoder, ExtentionType.DATA),  # feather
    "csv": (csv_decoder, ExtentionType.DATA),
    "jsonl": (jsonl_decoder, ExtentionType.DATA),
    "orc": (orc_decoder, ExtentionType.DATA),
    "parquet": (parquet_decoder, ExtentionType.DATA),
    "tsv": (tsv_decoder, ExtentionType.DATA),
    "zstd": (zstd_decoder, ExtentionType.DATA),  # jsonl/zstd
}



from opteryx.components.v2.logical_planner.planner import QUERY_BUILDERS


def engine_version(query):
    statement_type = next(iter(query))
    if statement_type in QUERY_BUILDERS:
        return "v2"
    return "v1"



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
No Operation

This is a SQL Query Execution Plan Node.
"""
from typing import Iterable

from opteryx.exceptions import SqlError
from opteryx.models import QueryProperties
from opteryx.operators import BasePlanNode


class NoOpNode(BasePlanNode):
    def __init__(self, properties: QueryProperties, **config):
        super().__init__(properties=properties)

    @property
    def name(self):  # pragma: no cover
        return "NoOp"

    @property
    def config(self):  # pragma: no cover
        return ""

    def execute(self) -> Iterable:
        yield None



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Heap Sort Node

This is a SQL Query Execution Plan Node.

This node orders a dataset, note that Heap Sort in this instance isn't the heap sort
algorithm, it is an approach where a heap of n items (the limit) is maintained as the
data passes through the operator. Because we are working with chunks, we build a 
temporary batch of n + len(batch) items, sort them and then discard all but n of
the items.

This is faster, particularly when working with large datasets even though we're now
sorting smaller datasets over and over again.

Early testing with a ~10m record dataset:
- sort then limit: 39.2 seconds
- heapsort: 9.8 seconds
"""
import time
from typing import Iterable

import pyarrow
from pyarrow import Table

from opteryx.exceptions import ColumnNotFoundError
from opteryx.exceptions import SqlError
from opteryx.managers.expression import NodeType
from opteryx.managers.expression import format_expression
from opteryx.models import Columns
from opteryx.models import QueryProperties
from opteryx.operators import BasePlanNode


class HeapSortNode(BasePlanNode):
    def __init__(self, properties: QueryProperties, **config):
        super().__init__(properties=properties)
        self.order = config.get("order", [])
        self.limit: int = config.get("limit", -1)

    @property
    def greedy(self):  # pragma: no cover
        return True

    @property
    def config(self):  # pragma: no cover
        return f"LIMIT = {self.limit} ORDER = " + ", ".join(
            [f"{i[0][0].value} {i[1][0:3].upper()}" for i in self.order]
        )

    @property
    def name(self):  # pragma: no cover
        return "Heap Sort"

    def execute(self) -> Iterable:
        if len(self._producers) != 1:  # pragma: no cover
            raise SqlError(f"{self.name} on expects a single producer")

        morsels = self._producers[0]  # type:ignore
        if isinstance(morsels, Table):
            morsels = (morsels,)

        columns = None
        collected_rows = None
        mapped_order = []

        for morsel in morsels.execute():
            if columns is None:
                columns = Columns(morsel)

                for column_list, direction in self.order:
                    for column in column_list:
                        if column.token_type == NodeType.LITERAL_NUMERIC:
                            # we have an index rather than a column name, it's a natural
                            # number but the list of column names is zero-based, so we
                            # subtract one
                            column_name = morsel.column_names[int(column.value) - 1]
                            mapped_order.append(
                                (
                                    column_name,
                                    direction,
                                )
                            )
                        else:
                            try:
                                mapped_order.append(
                                    (
                                        columns.get_column_from_alias(
                                            format_expression(column), only_one=True
                                        ),
                                        direction,
                                    )
                                )
                            except ColumnNotFoundError as cnfe:
                                raise ColumnNotFoundError(
                                    f"`ORDER BY` must reference columns as they appear in the `SELECT` clause. {cnfe}"
                                )

            start_time = time.time_ns()

            # add what we've collected before to the table
            if collected_rows:  # pragma: no cover
                self.statistics.morsel_merges += 1
                morsel = pyarrow.concat_tables([collected_rows, morsel], promote=True)

            morsel = morsel.sort_by(mapped_order)
            collected_rows = morsel.slice(offset=0, length=self.limit)

            self.statistics.time_ordering += time.time_ns() - start_time

        yield collected_rows



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Arrow Reader

Used to read datasets registered using the register_arrow or register_df functions.
"""

import typing

import pyarrow

from opteryx import config
from opteryx.connectors import BaseDocumentStorageAdapter
from opteryx.shared import MaterializedDatasets


class ArrowConnector(BaseDocumentStorageAdapter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._datasets = MaterializedDatasets()

    def get_document_count(self, collection) -> int:
        dataset = self._datasets[collection]
        return dataset.num_rows

    def read_documents(self, collection, morsel_size: typing.Union[int, None] = None):
        if morsel_size is None:
            morsel_size = config.MORSEL_SIZE
        dataset = self._datasets[collection]

        batch_size = morsel_size // (dataset.nbytes / dataset.num_rows)

        for batch in dataset.to_batches(max_chunksize=batch_size):
            morsel = pyarrow.Table.from_batches([batch], schema=dataset.schema)
            yield morsel



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This is a long module, the intent is to have the functions defined in alphabetical order to
help find them
"""

import datetime
import inspect
import re
import sys
import typing
from enum import Enum
from enum import auto
from functools import cache

CAMEL_TO_SNAKE = re.compile(r"(?<!^)(?=[A-Z])")


def not_implemented(*args, **kwds):
    raise NotImplementedError("Function Not Implemented")


@cache
def get_functions():
    """so we don't have to manually maintain a list of functions, build it using introspection"""
    functions = {}
    members = inspect.getmembers(
        sys.modules[__name__],
        lambda member: inspect.isclass(member) and member.__module__ == __name__,
    )
    for function_name, function_implementation in members:
        if function_name[0] != "_":
            # python CamelCase to SQL SNAKE_CASE
            function_name = CAMEL_TO_SNAKE.sub("_", function_name).upper()
            if function_name.startswith("FUNCTION_"):
                function_name = function_name[9:]
                functions[function_name] = function_implementation
    return functions


class _FunctionStyle(Enum):
    # Aggregation functions accept a column of values and return a single value
    AGGREGATION = auto()
    # Elementwise functions accept a set of two or more columns of the same length and return
    # a column of the same length as the inputs
    ELEMENTWISE = auto()
    # Setwise functions accept one column and one or more fixed values and return a column of
    # the same length as the input
    SETWISE = auto()
    # Constant functions return a single value, usually with no input
    CONSTANT = auto()


class _BaseFunction:
    # The inner function that is executed
    _func = not_implemented
    # The approximate cost to perform the function.
    # This is approximately the time to execute the function 1 million times in 1000s of a second.
    cost: typing.Union[int, None] = None
    # If this function can ever return a 'null' value
    returns_nulls: typing.Union[bool, None] = None
    # The range of values returnable by this value
    value_range: typing.Union[typing.Tuple[typing.Any], None] = None
    # The type of function
    style: _FunctionStyle = _FunctionStyle.ELEMENTWISE

    def __call__(self, *args: typing.Any, **kwds: typing.Any) -> typing.Any:
        return self._func(*args, **kwds)

    def __str__(self):
        return f"{self.__class__.__name__.upper()} (<params>) → <type>\n{self.__doc__.strip()}"


class FunctionCurrentTime(_BaseFunction):
    """Return the current system time."""

    style = _FunctionStyle.CONSTANT
    cost = 1

    def _func(self) -> datetime.time:
        return datetime.datetime.utcnow().time()


class FunctionE(_BaseFunction):
    """Return Euler's number."""

    style = _FunctionStyle.CONSTANT
    cost = 1

    def _func(self) -> float:
        return 2.71828182845904523536028747135266249775724709369995


class FunctionLen(_BaseFunction):
    """return the length of a VARCHAR or ARRAY"""

    style = _FunctionStyle.ELEMENTWISE

    def _func(self, item: typing.Union[list, str]) -> int:
        return len(item)


class FunctionPhi(_BaseFunction):
    """Return the golden ratio."""

    style = _FunctionStyle.CONSTANT
    cost = 1

    def _func(self) -> float:
        return 1.61803398874989484820458683436563811772030917980576


class FunctionPi(_BaseFunction):
    """Return Pi."""

    style = _FunctionStyle.CONSTANT
    cost = 1

    def _func(self) -> float:
        return 3.14159265358979323846264338327950288419716939937510


class FunctionVersion(_BaseFunction):
    """Return the version of the query engine."""

    style = _FunctionStyle.CONSTANT
    cost = 1

    def _func(self) -> str:
        import opteryx

        return opteryx.__version__


FUNCTIONS = get_functions()

if __name__ == "__main__":  # pragma: no cover
    func = FUNCTIONS["E"]()
    print(func)
    print(func())
    print(func.style)



import re


def tokenize_string(string):
    # Define the regular expression to match tokens
    token_pattern = re.compile(r"[(),]|[^\s(),]+")

    # Use the regular expression to find all tokens in the string
    tokens = token_pattern.findall(string)

    # Remove any leading or trailing whitespace from the tokens
    tokens = [token.strip() for token in tokens]

    return tokens


def format_sql(sql):
    # Split the SQL statement into individual words
    # this is not meant to cover all scenarios, it's mainly for debugging
    words = tokenize_string(sql)

    # Add spaces and new lines to format the SQL statement
    formatted_sql = ""
    indent_level = 0
    for word in words:
        if word == "(":
            indent_level += 1
            formatted_sql += "\033[38;5;102m" + word + "\033[0m "
        elif word == ")":
            formatted_sql += "\033[38;5;102m" + word + "\033[0m "
            indent_level -= 1
        elif word in [
            "SELECT",
            "FROM",
            "WHERE",
            "JOIN",
            "ON",
            "AND",
            "OR",
            "GROUP",
            "BY",
            "ORDER",
            "LIMIT",
        ]:
            formatted_sql += "\033[38;5;117m\n" + word.rjust(7) + "\033[0m "
        elif word in ["=", ">=", "<=", "!=", "<", ">", "<>"]:
            formatted_sql += "\033[38;5;183m" + word + "\033[0m "
        elif word.isdigit():
            formatted_sql += "\033[0;31m" + word + "\033[0m "
        else:
            formatted_sql += word + " "

    return "\n" + formatted_sql.strip()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Build Statistics Node

This is a SQL Query Execution Plan Node.

Gives information about a dataset's columns
"""
from typing import Iterable

from opteryx.exceptions import SqlError
from opteryx.models import QueryProperties
from opteryx.operators import BasePlanNode


class BuildStatisticsNode(BasePlanNode):
    def __init__(self, properties: QueryProperties, **config):
        super().__init__(properties=properties)

    @property
    def name(self):  # pragma: no cover
        return "Analyze Table"

    @property
    def config(self):  # pragma: no cover
        return ""

    def execute(self) -> Iterable:
        if len(self._producers) != 1:  # pragma: no cover  # pragma: no cover
            raise SqlError(f"{self.name} on expects a single producer")

        morsels = self._producers[0]  # type:ignore

        if morsels is None:  # pragma: no cover
            return None

        import orso

        for morsel in morsels.execute():
            from orso import converters

            from opteryx import utils

            rows, schema = converters.from_arrow(utils.arrow.rename_columns([morsel]))
            df = orso.DataFrame(rows=rows, schema=schema)
            yield df.profile.arrow()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import os
from typing import Optional

from opteryx.connectors import BaseBlobStorageAdapter
from opteryx.exceptions import MissingDependencyError
from opteryx.utils import paths


class GcpCloudStorageConnector(BaseBlobStorageAdapter):
    def __init__(self, project: Optional[str] = None, credentials=None, **kwargs):
        super().__init__(**kwargs)
        try:
            from google.auth.credentials import AnonymousCredentials
            from google.cloud import storage
        except ImportError as err:
            raise MissingDependencyError(err.name) from err

        #        super().__init__(**kwargs)
        self.project = project
        self.credentials = credentials

    def read_blob(self, blob_name):
        bucket, object_path, name, extension = paths.get_parts(blob_name)
        bucket = bucket.replace("va_data", "va-data")
        bucket = bucket.replace("data_", "data-")

        blob = get_blob(
            project=self.project,
            bucket=bucket,
            blob_name=object_path + name + extension,
        )
        stream = blob.download_as_bytes()
        return io.BytesIO(stream)

    def get_blob_list(self, partition=None):
        from google.cloud import storage

        bucket, object_path, name, extension = paths.get_parts(partition)
        bucket = bucket.replace("va_data", "va-data")
        bucket = bucket.replace("data_", "data-")

        # this means we're not actually going to GCP
        if os.environ.get("STORAGE_EMULATOR_HOST") is not None:
            from google.auth.credentials import AnonymousCredentials

            client = storage.Client(
                credentials=AnonymousCredentials(),
                project=self.project,
            )
        else:  # pragma: no cover
            client = storage.Client(project=self.project)

        gcs_bucket = client.get_bucket(bucket)
        blobs = list(client.list_blobs(bucket_or_name=gcs_bucket, prefix=object_path))

        return [bucket + "/" + blob.name for blob in blobs if not blob.name.endswith("/")]


def get_blob(project: str, bucket: str, blob_name: str):
    from google.cloud import storage

    # this means we're not actually going to GCP
    if os.environ.get("STORAGE_EMULATOR_HOST") is not None:
        from google.auth.credentials import AnonymousCredentials

        client = storage.Client(
            credentials=AnonymousCredentials(),
            project=project,
        )
    else:  # pragma: no cover
        client = storage.Client(project=project)

    gcs_bucket = client.get_bucket(bucket)
    blob = gcs_bucket.get_blob(blob_name)
    return blob



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A MongoDB Reader
This is a light-weight MongoDB reader to fulfil a specific purpose,
it needs some work to make it fully reusable.

Based on the now deprecated Mabel MongoDB reader
https://github.com/mabel-dev/mabel/blob/6bcd978b90870187d5eff939be3f5845a3cdf900/mabel/adapters/mongo/mongodb_reader.py
"""
import os
import typing

from opteryx import config
from opteryx.connectors import BaseDocumentStorageAdapter
from opteryx.exceptions import UnmetRequirementError

try:
    import pymongo  # type:ignore
except ImportError:  # pragma: no cover
    pass


class MongoDbConnector(BaseDocumentStorageAdapter):
    def __init__(self, *args, **kwargs):
        BaseDocumentStorageAdapter.__init__(self, *args, **kwargs)
        """establish the connection to mongodb"""
        mongo_connection = os.environ.get("MONGO_CONNECTION")
        mongo_database = os.environ.get("MONGO_DATABASE")

        if mongo_connection is None or mongo_database is None:  # pragma: no cover
            raise UnmetRequirementError(
                "MongoDB adapter requires MONGO_CONNECTION and MONGO_DATABASE set in environment variables."
            )

        client = pymongo.MongoClient(mongo_connection)  # type:ignore
        self._database = client[mongo_database]

    def get_document_count(self, collection) -> int:
        """
        Return the count, or an estimate of, the number of documents
        """
        _collection = collection.split(".")[0]
        return self._database[_collection].estimated_document_count()

    def read_documents(self, collection, morsel_size: typing.Union[int, None] = None):
        """
        Return a morsel of documents
        """
        if morsel_size is None:
            morsel_size = config.MORSEL_SIZE

        _collection = collection.split(".")[0]
        documents = self._database[_collection].find()
        for morsel in self.chunk_dictset(documents, morsel_size):
            yield morsel



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy

from opteryx.exceptions import SqlError
from opteryx.utils import dates
from opteryx.utils import intervals


def generate_series(*args):
    from opteryx.managers.expression import NodeType

    arg_len = len(args)
    arg_vals = [i.value for i in args]
    first_arg_type = args[0].token_type

    # if the parameters are numbers, generate series is an alias for range
    if first_arg_type in (NodeType.LITERAL_NUMERIC, numpy.float64):
        if arg_len not in (2, 3):
            raise SqlError("generate_series for numbers takes 2 or 3 parameters.")
        return intervals.generate_range(*arg_vals)

    # if the params are timestamps, we create time intervals
    if first_arg_type == NodeType.LITERAL_TIMESTAMP:
        if arg_len != 3:
            raise SqlError("generate_series for dates needs start, end, and interval parameters")
        return dates.date_range(*arg_vals)

    raise SqlError("Unsupported value for GENERATE_SERIES")



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## bind variables
## bind temporal ranges
## (bind statistics)

import datetime
import decimal
from typing import Iterable

import numpy

import opteryx
from opteryx.exceptions import ProgrammingError
from opteryx.exceptions import SqlError
from opteryx.managers.expression import ExpressionTreeNode
from opteryx.models import QueryProperties


def _build_literal_node(value):
    if value is None:
        return {"Value": "Null"}
    if isinstance(value, (bool)):
        # boolean must be before numeric
        return {"Value": {"Boolean": value}}
    if isinstance(value, (str)):
        return {"Value": {"SingleQuotedString": value}}
    if isinstance(value, (int, float, decimal.Decimal)):
        return {"Value": {"Number": [value, False]}}
    if isinstance(value, (numpy.datetime64)):
        return {"Value": {"SingleQuotedString": value.item().isoformat()}}
    if isinstance(value, (datetime.date, datetime.datetime)):
        return {"Value": {"SingleQuotedString": value.isoformat()}}


def variable_binder(node, parameter_set, properties, query_type):
    """Walk the AST replacing 'Placeholder' nodes, this is recursive"""
    # Replace placeholders with parameters.
    # We do this after the AST has been parsed to remove any chance of the
    # parameter affecting the meaning of any of the other tokens - i.e. to
    # eliminate this feature being used for SQL injection.
    if isinstance(node, list):
        return [variable_binder(i, parameter_set, properties, query_type) for i in node]
    if isinstance(node, dict):
        if "Value" in node:
            if "Placeholder" in node["Value"]:
                # fmt:off
                if len(parameter_set) == 0:
                    raise ProgrammingError("Incorrect number of bindings supplied."
                    " More placeholders are provided than parameters.")
                placeholder_value = parameter_set.pop(0)
                # prepared statements will have parsed this already
                if isinstance(placeholder_value, ExpressionTreeNode):
                    placeholder_value = placeholder_value.value
                return _build_literal_node(placeholder_value)
                # fmt:on
        # replace @variables
        if query_type != "SetVariable" and "Identifier" in node:
            token_name = node["Identifier"]["value"]
            if token_name.startswith("@@"):
                return _build_literal_node(opteryx.variables.get(token_name[2:]))
            elif token_name[0] == "@":
                if token_name not in properties.variables:  # pragma: no cover
                    raise SqlError(f"Undefined variable found in query `{token_name}`.")
                variable_value = properties.variables.get(token_name)
                return _build_literal_node(variable_value.value)
        return {
            k: variable_binder(v, parameter_set, properties, query_type) for k, v in node.items()
        }
    # we're a leaf
    return node


def temporal_range_binder(ast, filters, cache):
    if isinstance(ast, (list)):
        return [temporal_range_binder(node, filters, cache) for node in ast]
    if isinstance(ast, (dict)):
        node_name = next(iter(ast))
        if node_name == "Table":
            temporal_range = filters.pop(0)
            ast["Table"]["start_date"] = temporal_range[1]
            ast["Table"]["end_date"] = temporal_range[2]
            ast["Table"]["cache"] = cache
            return ast
        if "table_name" in ast:
            temporal_range = filters.pop(0)
            ast["table_name"][0]["start_date"] = temporal_range[1]
            ast["table_name"][0]["end_date"] = temporal_range[2]
            ast["table_name"][0]["cache"] = cache
            return ast
        if "ShowCreate" in ast:
            temporal_range = filters.pop(0)
            ast["ShowCreate"]["start_date"] = temporal_range[1]
            ast["ShowCreate"]["end_date"] = temporal_range[2]
            ast["ShowCreate"]["cache"] = cache
            return ast
        return {k: temporal_range_binder(v, filters, cache) for k, v in ast.items()}
    return ast


def statistics_binder(ast):
    return ast


def bind_ast(ast, parameters: Iterable = None, properties: QueryProperties = None):
    """
    Bind physical information to the AST

    This includes the following activities
    - Replacing placeholders with the parameters
    - Adding temporal range information to relations
    """
    query_type = next(iter(ast))

    bound_ast = ast.copy()
    bound_ast = variable_binder(bound_ast, parameters, properties, query_type)
    if query_type in ("Analyze", "Query", "Explain", "ShowColumns", "ShowCreate"):
        bound_ast = temporal_range_binder(
            bound_ast, list(properties.temporal_filters), properties.cache
        )
        bound_ast = statistics_binder(bound_ast)

    return bound_ast



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing

from opteryx.constants.character_set import CharacterSet
from opteryx.constants.character_set import Collation
from opteryx.exceptions import VariableNotFoundError

VariableSchema = typing.Tuple[typing.Type, typing.Any]

SYSTEM_VARIABLES: typing.Dict[str, VariableSchema] = {
    # name: (type, default)
    "auto_increment_increment": (int, 1),
    "autocommit": (bool, True),
    "character_set_client": (str, CharacterSet.utf8mb4.name),
    "character_set_connection": (str, CharacterSet.utf8mb4.name),
    "character_set_database": (str, CharacterSet.utf8mb4.name),
    "character_set_results": (str, CharacterSet.utf8mb4.name),
    "character_set_server": (str, CharacterSet.utf8mb4.name),
    "collation_connection": (str, Collation.utf8mb4_general_ci.name),
    "collation_database": (str, Collation.utf8mb4_general_ci.name),
    "collation_server": (str, Collation.utf8mb4_general_ci.name),
    "external_user": (str, ""),
    "init_connect": (str, ""),
    "interactive_timeout": (int, 28800),
    "license": (str, "MIT"),
    "lower_case_table_names": (int, 0),
    "max_allowed_packet": (int, 67108864),
    "max_execution_time": (int, 0),
    "net_buffer_length": (int, 16384),
    "net_write_timeout": (int, 28800),
    "performance_schema": (bool, False),
    "sql_auto_is_null": (bool, False),
    "sql_mode": (str, "ANSI"),
    "sql_select_limit": (int, None),
    "system_time_zone": (str, "UTC"),
    "time_zone": (str, "UTC"),
    "transaction_read_only": (bool, False),
    "transaction_isolation": (str, "READ-COMMITTED"),
    "version": (str, "8.0.29"),
    "version_comment": (str, "mesos"),
    "wait_timeout": (int, 28800),
    "event_scheduler": (str, "OFF"),
    "default_storage_engine": (str, "opteryx"),
    "default_tmp_storage_engine": (str, "opteryx"),
}


def get(name):
    if not name in SYSTEM_VARIABLES:
        raise VariableNotFoundError(name)
    return SYSTEM_VARIABLES[name][1]


def assign(name: str, value: typing.Union[str, int, bool]):
    if not name in SYSTEM_VARIABLES:
        raise VariableNotFoundError(name)
    variable_type, _ = SYSTEM_VARIABLES[name]
    if not isinstance(value, variable_type):
        raise ValueError(f"Invalid type for {name}.")
    SYSTEM_VARIABLES[name] = (variable_type, value)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

import orjson


class Node:
    def _is_valid_key(self, key):
        if key.startswith("_"):
            raise AttributeError("Node cannot have dynamic attributes starting with an underscore")

    def __init__(self, *args, **kwargs):
        """
        A Python object with run-time defined attributes.
        """
        internal = {}
        if len(args) == 1:
            internal["node_type"] = str(args[0])
        if len(args) > 1:
            raise ValueError("Only one position argument can be passed to a Node initializer")
        if isinstance(kwargs, dict):
            for k, v in kwargs.items():
                self._is_valid_key(k)
                internal[k] = v
        self.__dict__["_internal"] = internal

    def __getattr__(self, __name: str) -> Any:
        internal = self.__dict__.get("_internal", {})
        return internal.get(__name)

    def __setattr__(self, __name: str, __value: Any) -> None:
        internal = self.__dict__.get("_internal", {})
        if __name != "_internal":
            self._is_valid_key(__name)
            if __value is None:
                internal.pop(__name, None)
            else:
                internal[__name] = __value

    def __str__(self) -> str:
        internal = self.__dict__.get("_internal", {})
        return orjson.dumps(internal).decode()

    def __repr__(self) -> str:
        node_type = str(self.node_type)
        if node_type.startswith("LogicalPlanStepType."):
            node_type = node_type[20:]
        node_params: list = []
        return node_type + " (" + ",".join(node_params) + ")"


if __name__ == "__main__":  # pragma: no cover
    n = Node(a=3)
    n.name = "john"
    print(n.a)
    print(n.b)
    print(n.name)
    print(repr(n))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Implement conditions which are essentially unary statements, usually IS statements.

This are executed as functions on arrays rather than functions on elements in arrays.
"""
import numpy
from pyarrow import compute


def _is_null(values):
    indicies = numpy.nonzero(compute.is_null(values, nan_is_null=True))[0]
    mask_array = numpy.zeros(len(values), dtype=bool)
    mask_array[indicies] = True
    return mask_array


def _is_not_null(values):
    indicies = compute.is_null(values, nan_is_null=True)
    indicies = numpy.nonzero(numpy.invert(indicies))[0]
    mask_array = numpy.zeros(len(values), dtype=bool)
    mask_array[indicies] = True
    return mask_array


def _is_true(values):
    indicies = numpy.nonzero(values)[0]
    mask_array = numpy.zeros(len(values), dtype=bool)
    mask_array[indicies] = True
    return mask_array


def _is_false(values):
    indicies = numpy.invert(values)
    indicies = numpy.nonzero(indicies)[0]
    mask_array = numpy.zeros(len(values), dtype=bool)
    mask_array[indicies] = True
    return mask_array


UNARY_OPERATIONS = {
    "IsNull": _is_null,
    "IsNotNull": _is_not_null,
    "IsTrue": _is_true,
    "IsFalse": _is_false,
}



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Represents a logical plan

A small set of functions are available in the logical plan (a set similar to, but
different from Cobb's relational algebra)

Steps are given random IDs to prevent collisions
"""

"""1xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"""
import os
import sys
from enum import Enum
from enum import auto

from opteryx.components.logical_planner import builders
from opteryx.managers.expression import ExpressionTreeNode
from opteryx.managers.expression import NodeType
from opteryx.managers.expression import get_all_nodes_of_type
from opteryx.models.node import Node
from opteryx.third_party.travers import Graph
from opteryx.utils import random_string

sys.path.insert(1, os.path.join(sys.path[0], "../../../.."))


"""2xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"""


class LogicalPlanStepType(int, Enum):
    Project = auto()  # field selection
    Filter = auto()  # tuple filtering
    Union = auto()  #  appending relations
    Difference = auto()  # relation interection
    Join = auto()  # all joina
    Group = auto()  # group by, without the aggregation
    Aggregate = auto()
    Scan = auto()  # read a dataset
    Set = auto()  # set a variable
    Limit = auto()  # limit and offset
    Order = auto()  # order by
    Distinct = auto()

    CTE = auto()
    Subquery = auto()


class LogicalPlan(Graph):
    pass


"""
CLAUSE PLANNERS
"""


def extract_ctes(branch, planner):
    ctes = {}
    if branch["with"]:
        for _ast in branch["with"]["cte_tables"]:
            alias = _ast.get("alias")["name"]["value"]
            plan = {"Query": _ast["query"]}
            ctes[alias] = planner(plan)
    return ctes


def extract_value(clause):
    if len(clause) == 1:
        return builders.build(clause[0])
    return [builders.build(token) for token in clause]


def extract_variable(clause):
    if len(clause) == 1:
        return clause[0]["value"]
    return [token["value"] for token in clause]


"""
STATEMENT PLANNERS
"""


def plan_query(statement):
    """ """

    def _inner_query_planner(sub_plan):
        inner_plan = LogicalPlan()
        step_id = None

        # from
        _relations = sub_plan["Select"]["from"]
        for relation in _relations:
            if "Derived" in relation["relation"]:
                if relation["relation"]["Derived"]["subquery"]:
                    subquery_step = Node(node_type=LogicalPlanStepType.Subquery)
                    previous_step_id, step_id = step_id, random_string()
                    previous_from_step_id, from_step_id = previous_step_id, step_id
                    if previous_step_id is not None:
                        inner_plan.add_edge(previous_step_id, step_id)
                    inner_plan.add_node(from_step_id, subquery_step)

                    subquery_plan = plan_query(relation["relation"]["Derived"]["subquery"])
                    inner_plan += subquery_plan
                    subquery_entry_id = subquery_plan.get_entry_points()[0]
                    inner_plan.add_edge(from_step_id, subquery_entry_id)

                else:
                    raise NotImplementedError(relation["relation"]["Derived"])
            else:
                from_step = Node(node_type=LogicalPlanStepType.Scan, relation=relation)
                previous_step_id, step_id = step_id, random_string()
                previous_from_step_id, from_step_id = previous_step_id, step_id
                inner_plan.add_node(from_step_id, from_step)

            # joins
            _joins = relation["joins"]
            for join in _joins:
                # add the join node
                join_step = Node(node_type=LogicalPlanStepType.Join, join=join)
                previous_step_id, step_id = step_id, random_string()
                join_step_id = step_id
                inner_plan.add_node(join_step_id, join_step)
                # add the from table as the left side of the join
                inner_plan.add_edge(from_step_id, join_step_id, "left")
                # add the other side of the join
                # TODO: if it's a subquery, expand it out
                right_node = random_string()
                joined_read_step = Node(
                    node_type=LogicalPlanStepType.Scan, relation=join["relation"]
                )
                inner_plan.add_node(right_node, joined_read_step)
                inner_plan.add_edge(right_node, join_step_id, "right")

            if len(_joins) == 0:
                if previous_from_step_id is not None:
                    inner_plan.add_edge(previous_from_step_id, step_id)

        # TODO: if there's any orphaned relations, they are implicit cross joins

        # selection
        _selection = builders.build(sub_plan["Select"]["selection"])
        if _selection:
            selection_step = Node(node_type=LogicalPlanStepType.Filter)
            previous_step_id, step_id = step_id, random_string()
            inner_plan.add_node(step_id, selection_step)
            if previous_step_id is not None:
                inner_plan.add_edge(previous_step_id, step_id)

        # groups
        _groups = builders.build(sub_plan["Select"]["group_by"])
        if _groups != []:
            group_step = Node(node_type=LogicalPlanStepType.Group, group=_groups)
            previous_step_id, step_id = step_id, random_string()
            inner_plan.add_node(step_id, group_step)
            if previous_step_id is not None:
                inner_plan.add_edge(previous_step_id, step_id)

        # aggregates
        _projection = builders.build(sub_plan["Select"]["projection"])
        _aggregates = get_all_nodes_of_type(
            _projection, select_nodes=(NodeType.AGGREGATOR, NodeType.COMPLEX_AGGREGATOR)
        )
        if len(_aggregates) > 0:
            aggregate_step = Node(node_type=LogicalPlanStepType.Aggregate, aggregates=_aggregates)
            previous_step_id, step_id = step_id, random_string()
            inner_plan.add_node(step_id, aggregate_step)
            if previous_step_id is not None:
                inner_plan.add_edge(previous_step_id, step_id)

        # projection
        _projection = [clause for clause in _projection if clause not in _aggregates]
        if not (len(_projection) == 1 and _projection[0].token_type == NodeType.WILDCARD):
            project_step = Node(node_type=LogicalPlanStepType.Project, projection=_projection)
            previous_step_id, step_id = step_id, random_string()
            inner_plan.add_node(step_id, project_step)
            if previous_step_id is not None:
                inner_plan.add_edge(previous_step_id, step_id)

        # having
        _having = builders.build(sub_plan["Select"]["having"])
        if _having:
            having_step = Node(node_type=LogicalPlanStepType.Filter)
            previous_step_id, step_id = step_id, random_string()
            inner_plan.add_node(step_id, having_step)
            if previous_step_id is not None:
                inner_plan.add_edge(previous_step_id, step_id)

        # distinct
        if sub_plan["Select"]["distinct"]:
            distinct_step = Node(node_type=LogicalPlanStepType.Distinct)
            previous_step_id, step_id = step_id, random_string()
            inner_plan.add_node(step_id, distinct_step)
            if previous_step_id is not None:
                inner_plan.add_edge(previous_step_id, step_id)

        # order
        _order_by = sub_plan["order_by"]
        if _order_by:
            order_step = Node(node_type=LogicalPlanStepType.Order, order=_order_by)
            previous_step_id, step_id = step_id, random_string()
            inner_plan.add_node(step_id, order_step)
            if previous_step_id is not None:
                inner_plan.add_edge(previous_step_id, step_id)

        # limit/offset
        _limit = sub_plan["limit"]
        _offset = sub_plan["offset"]
        if _limit or _offset:
            limit_step = Node(node_type=LogicalPlanStepType.Limit, limit=_limit, offset=_offset)
            previous_step_id, step_id = step_id, random_string()
            inner_plan.add_node(step_id, limit_step)
            if previous_step_id is not None:
                inner_plan.add_edge(previous_step_id, step_id)

        return inner_plan

    root_node = statement
    if "Query" in root_node:
        root_node = root_node["Query"]

    # CTEs need to be extracted so we can deal with them later
    raw_ctes = extract_ctes(root_node, _inner_query_planner)

    # union?
    if "SetOperator" in root_node["body"]:
        plan = LogicalPlan()
        root_node = root_node["body"]["SetOperator"]
        _left = _inner_query_planner(root_node["left"])
        _right = _inner_query_planner(root_node["right"])
        _operator = root_node["op"]
        # join the plans together
        raise NotImplementedError("Set Operators (UNION) not implemented")

    # we do some minor AST rewriting
    root_node["body"]["limit"] = root_node.get("limit", None)
    root_node["body"]["offset"] = root_node.get("offset", None)
    root_node["body"]["order_by"] = root_node.get("order_by", None)
    return _inner_query_planner(root_node["body"])


def plan_set_variable(statement):
    root_node = "SetVariable"
    plan = LogicalPlan()
    set_step = Node(
        node_type=LogicalPlanStepType.Set,
        variable=extract_variable(statement[root_node]["variable"]),
        value=extract_value(statement[root_node]["value"]),
    )
    plan.add_node(random_string(), set_step)
    return plan


def plan_show_variables(statement):
    root_node = "ShowVariables"
    plan = LogicalPlan()

    read_step = Node(node_type=LogicalPlanStepType.Scan, source="$variables")
    step_id = random_string()
    plan.add_node(step_id, read_step)

    predicate = statement[root_node]["filter"]
    if predicate is not None:
        operator = next(iter(predicate))
        select_step = Node(
            node_type=LogicalPlanStepType.Filter,
            predicate=ExpressionTreeNode(
                token_type=NodeType.COMPARISON_OPERATOR,
                value=operator,
                left=ExpressionTreeNode(token_type=NodeType.IDENTIFIER, value="name"),
                right=predicate[operator],
            ),
        )
        previous_step_id, step_id = step_id, random_string()
        plan.add_node(step_id, select_step)
        plan.add_edge(previous_step_id, step_id)

    return plan


QUERY_BUILDERS = {
    #    "Analyze": analyze_query,
    #    "Explain": explain_query,
    "Query": plan_query,
    "SetVariable": plan_set_variable,
    #    "ShowColumns": show_columns_query,
    #    "ShowCreate": show_create_query,
    #    "ShowFunctions": show_functions_query,
    #    "ShowVariable": show_variable_query,  # generic SHOW handler
    "ShowVariables": plan_show_variables,
}


def get_planners(parsed_statements):
    # The sqlparser ast is an array of asts
    for parsed_statement in parsed_statements:
        statement_type = next(iter(parsed_statement))
        yield QUERY_BUILDERS[statement_type], parsed_statement


def print_tree(tree, lengths=None):
    """
    Prints a nested dictionary as an ascii tree
    """
    if lengths is None:
        lengths = {}

    name = tree["name"]
    depth = tree["depth"]
    relationship = tree.get("relationship")

    # Calculate the length of the name and relationship strings
    name_length = len(name)
    rel_length = len(relationship) if relationship is not None else 0

    # Update the maximum length for this depth if necessary
    if depth not in lengths:
        lengths[depth] = {"name": name_length, "relationship": rel_length}
    else:
        lengths[depth]["name"] = max(lengths[depth]["name"], name_length)
        lengths[depth]["relationship"] = max(lengths[depth]["relationship"], rel_length)

    # Print the node
    print(" " * (depth * 4), end="")
    print(tree["type"], end="")
    if relationship is not None:
        print(" (%s)" % relationship, end="")
    print()

    # Recursively print the children
    for child in tree["children"]:
        print_tree(child, lengths)

    # Print a blank line to separate nodes at the same depth
    if depth == 0:
        print()

    # Return the lengths for the parent to use
    return lengths


if __name__ == "__main__":  # pragma: no cover
    import json

    import opteryx.third_party.sqloxide

    SQL = "SET enable_optimizer = 7"
    TSQL = "SELECT * FROM $planets"
    TSQL = "SELECT DISTINCT MAX(planetId), name FROM $satellites INNER JOIN $planets ON $planets.id = $satellites.id WHERE id = 1 GROUP BY planetId HAVING id > 2 ORDER BY name LIMIT 1 OFFSET 1"
    TSQL = "SELECT * FROM T1, T2"
    TSQL = "SET @planet = 'Saturn'; SELECT name AS nom FROM (SELECT DISTINCT id as planetId, name FROM $planets WHERE name = @planet) as planets -- LEFT JOIN (SELECT planetId, COUNT(*) FROM $satellites FOR DATES BETWEEN '2022-01-01' AND TODAY WHERE gm > 10) AS bigsats ON bigsats.planetId = planets.planetId -- LEFT JOIN (SELECT planetId, COUNT(*) FROM $satellites FOR DATES IN LAST_MONTH WHERE gm < 10) as smallsats ON smallsats.planetId = planets.planetId ; "
    SQL = """SELECT name AS nom , bigsats.occurances , smallsats.occurances 
  FROM ( SELECT DISTINCT id as planetId , name FROM $planets WHERE name = @planet ) as planets 
  LEFT JOIN ( SELECT planetId , COUNT ( * ) AS occurances FROM $satellites WHERE gm > 10 GROUP BY planetId ) AS bigsats 
       ON bigsats.planetId = planets.planetId
  LEFT JOIN ( SELECT planetId , COUNT ( * ) AS occurances FROM $satellites WHERE gm < 10 GROUP BY planetId ) as smallsats 
       ON smallsats.planetId = planets.planetId ;"""
    TSQL = "SELECT COUNT(*) FROM $astronauts WHERE $astronauts.a = $astronauts.b"

    parsed_statements = opteryx.third_party.sqloxide.parse_sql(SQL, dialect="mysql")
    print(json.dumps(parsed_statements, indent=2))
    for planner, ast in get_planners(parsed_statements):
        print("---")
        tree = planner(ast)

        print(tree.breadth_first_search(tree.get_entry_points()[0]))
        print(tree.draw())
        print(json.dumps(tree.depth_first_search(), indent=2))
        print(print_tree(tree.depth_first_search()))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Show Functions Node

This is a SQL Query Execution Plan Node.
"""
from typing import Iterable

import pyarrow

from opteryx import functions
from opteryx import operators
from opteryx.models import Columns
from opteryx.models import QueryProperties
from opteryx.operators import BasePlanNode


class ShowFunctionsNode(BasePlanNode):
    def __init__(self, properties: QueryProperties, **config):
        super().__init__(properties=properties)

    @property
    def name(self):  # pragma: no cover
        return "Show Functions"

    @property
    def config(self):  # pragma: no cover
        return ""

    def execute(self) -> Iterable:
        buffer = []

        for function in functions.functions():
            buffer.append({"name": function, "type": "function"})
        for aggregate in operators.aggregators():
            buffer.append({"name": aggregate, "type": "aggregator"})

        table = pyarrow.Table.from_pylist(buffer)
        table = Columns.create_table_metadata(
            table=table,
            expected_rows=len(buffer),
            name="show_functions",
            table_aliases=[],
            disposition="calculated",
            path="show_functions",
        )

        yield table
        return



from typing import Any
from typing import Dict

import numpy
from Cython.Build import cythonize
from setuptools import Extension
from setuptools import find_packages
from setuptools import setup
from setuptools_rust import RustExtension

LIBRARY = "opteryx"


def rust_build(setup_kwargs: Dict[str, Any]) -> None:
    setup_kwargs.update(
        {
            "rust_extensions": [
                RustExtension("opteryx.third_party.sqloxide.sqloxide", "Cargo.toml", debug=False)
            ],
            "zip_safe": False,
        }
    )


__version__ = "notset"
with open(f"{LIBRARY}/version.py", mode="r") as v:
    vers = v.read()
exec(vers)  # nosec

with open("README.md", mode="r", encoding="UTF8") as rm:
    long_description = rm.read()

try:
    with open("requirements.txt", "r") as f:
        required = f.read().splitlines()
except:
    with open(f"{LIBRARY}.egg-info/requires.txt", "r") as f:
        required = f.read().splitlines()

extensions = [
    Extension(
        name="cjoin",
        sources=["opteryx/third_party/pyarrow_ops/cjoin.pyx"],
        include_dirs=[numpy.get_include()],
    ),
    Extension(
        name="csoundex",
        sources=["opteryx/third_party/fuzzy/csoundex.pyx"],
        include_dirs=[numpy.get_include()],
    ),
    Extension(
        name="clevenshtein",
        sources=["opteryx/third_party/levenshtein/clevenshtein.pyx"],
        include_dirs=[numpy.get_include()],
    ),
]

setup_config = {
    "name": LIBRARY,
    "version": __version__,
    "description": "Python SQL Query Engine",
    "long_description": long_description,
    "long_description_content_type": "text/markdown",
    "maintainer": "@joocer",
    "author": "@joocer",
    "author_email": "justin.joyce@joocer.com",
    "packages": find_packages(include=[LIBRARY, f"{LIBRARY}.*"]),
    "python_requires": ">=3.8",
    "url": "https://github.com/mabel-dev/opteryx/",
    "install_requires": required,
    "ext_modules": cythonize(extensions),
    "entry_points": {
        "console_scripts": ["opteryx=opteryx.command:main"],
    },
}

rust_build(setup_config)

setup(**setup_config)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
File Reader Node

This is a SQL Query Execution Plan Node.

This Node reads files to a PyArrow Table.
"""
import time
from typing import Iterable

import pyarrow

from opteryx.managers.expression import ExpressionTreeNode
from opteryx.managers.expression import NodeType
from opteryx.models import Columns
from opteryx.models import QueryProperties
from opteryx.operators import BasePlanNode
from opteryx.utils.file_decoders import KNOWN_EXTENSIONS


def _normalize_to_types(table):
    """
    Normalize types e.g. all numbers are decimal128 and dates
    """
    schema = table.schema

    for index, column_name in enumerate(schema.names):
        type_name = str(schema.types[index])
        if type_name in ("date32[day]", "date64", "timestamp[s]", "timestamp[ms]"):
            schema = schema.set(
                index,
                pyarrow.field(
                    name=column_name,
                    type=pyarrow.timestamp("us"),
                    metadata=table.field(column_name).metadata,
                ),
            )
        if type_name == ("list<item: null>"):
            schema = schema.set(
                index,
                pyarrow.field(
                    name=column_name,
                    type=pyarrow.list_(pyarrow.string()),
                    metadata=table.field(column_name).metadata,
                ),
            )

    return table.cast(target_schema=schema), schema


class FileReaderNode(BasePlanNode):
    def __init__(self, properties: QueryProperties, **config):
        """
        The Blob Reader Node is responsible for reading the relevant blobs
        and returning a Table/Relation.
        """
        super().__init__(properties=properties)

        self._dataset: str = config.get("dataset", None)
        self._alias: list = config.get("alias", None)

        self._filter = None

        self._reader = config.get("reader")  # type:ignore

        # pushed down projections
        self._selection = config.get("selection")
        if isinstance(self._selection, list):
            self._selection = set(self._selection)

    @property
    def can_push_selection(self):
        return True

    def push_predicate(self, predicate):
        # For the blob reader, we push selection nodes, for parquet we then convert
        # these to DNF at read time, for everything else, we run the selection nodes
        from opteryx.connectors.capabilities import predicate_pushable

        # Appears to be a bug in how pyarrow does indentifier filters so don't push these
        if predicate.left.token_type == predicate.right.token_type == NodeType.IDENTIFIER:
            return False
        if predicate_pushable.to_dnf(predicate) is None:
            # we can't push all predicates everywhere
            return False
        if self._filter is None:
            self._filter = predicate
            return True
        self._filter = ExpressionTreeNode(NodeType.AND, left=predicate, right=self._filter)
        return True

    @property
    def config(self):  # pragma: no cover
        if self._alias:
            return f"{self._dataset} => {self._alias}"
        if isinstance(self._dataset, str):
            return f"{self._dataset}"
        return "<complex dataset>"

    @property
    def name(self):  # pragma: no cover
        return "File Reader"

    def execute(self) -> Iterable:
        ext = self._dataset.split("/")[-1].split(".")[-1]
        parser, _ = KNOWN_EXTENSIONS[ext]

        time_to_read, blob_bytes, pyarrow_blob = self._read_and_parse(
            (
                self._dataset,
                self._reader.read_blob,
                parser,
                self._selection,
                self._filter,
            )
        )

        # we've read a blob
        self.statistics.count_data_blobs_read += 1

        # extract stats from reader
        self.statistics.bytes_read_data += blob_bytes
        self.statistics.time_data_read += time_to_read

        # we should know the number of entries
        self.statistics.rows_read += pyarrow_blob.num_rows
        self.statistics.bytes_processed_data += pyarrow_blob.nbytes

        pyarrow_blob = Columns.create_table_metadata(
            table=pyarrow_blob,
            expected_rows=0,
            name=self._dataset,
            table_aliases=[self._alias],
            disposition="blob",
            path=self._dataset,
        )
        self.statistics.columns_read += len(pyarrow_blob.column_names)

        pyarrow_blob, schema = _normalize_to_types(pyarrow_blob)

        # break up large files
        batch_size = (96 * 1024 * 1024) // (pyarrow_blob.nbytes / pyarrow_blob.num_rows)
        for batch in pyarrow_blob.to_batches(max_chunksize=batch_size):
            yield pyarrow.Table.from_batches([batch], schema=pyarrow_blob.schema)

    def _read_and_parse(self, config):
        path, reader, parser, projection, selection = config
        start_read = time.time_ns()
        blob_bytes = reader(path)
        table = parser(blob_bytes, projection, selection)
        time_to_read = time.time_ns() - start_read
        return time_to_read, blob_bytes.getbuffer().nbytes, table



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os

from orso.logging import get_logger
from orso.logging import set_log_name

from opteryx import config
from opteryx.connection import Connection
from opteryx.connectors import register_arrow
from opteryx.connectors import register_df
from opteryx.connectors import register_store
from opteryx.shared import variables
from opteryx.version import __version__

set_log_name("OPTERYX")
logger = get_logger()

__author__: str = "@joocer"

apilevel: str = "1.0"  # pylint: disable=C0103
threadsafety: int = 0  # pylint: disable=C0103
paramstyle: str = "qmark"  # pylint: disable=C0103


def connect(*args, **kwargs):
    """define the opteryx.connect function"""
    return Connection(*args, **kwargs)


def query(operation, *args, params: list = None, **kwargs):
    """helper routine, create a connection and return an executed cursor"""
    # query is the similar DuckDB function
    conn = Connection(*args, **kwargs)
    curr = conn.cursor()
    curr.execute(operation=operation, params=params)
    return curr


# Try to increase the priority of the application
if not config.DISABLE_HIGH_PRIORITY and hasattr(os, "nice"):  # pragma: no cover
    nice_value = os.nice(0)
    try:
        os.nice(-20 + nice_value)
        logger.info(f"Process priority set to {os.nice(0)}.")
    except PermissionError:
        display_nice = str(nice_value)
        if nice_value == 0:
            display_nice = "0 (normal)"
        logger.info(f"Cannot update process priority. Currently set to {display_nice}.")

# Log resource usage
if config.ENABLE_RESOURCE_LOGGING:  # pragma: no cover
    from opteryx.utils.resource_monitor import ResourceMonitor



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Cross Join Node

This is a SQL Query Execution Plan Node.

This performs a CROSS JOIN - CROSS JOIN is not natively supported by PyArrow so this is written
here rather than calling the join() functions
"""
import typing

import numpy
import pyarrow

from opteryx.exceptions import SqlError
from opteryx.managers.expression import NodeType
from opteryx.models import Columns
from opteryx.models import QueryProperties
from opteryx.operators import BasePlanNode

INTERNAL_BATCH_SIZE = 100  # config
MAX_JOIN_SIZE = 500  # config


def _cartesian_product(*arrays):
    """
    Cartesian product of arrays creates every combination of the elements in the arrays
    """
    array_count = len(arrays)
    arr = numpy.empty([len(array) for array in arrays] + [array_count], dtype=numpy.int64)
    for i, array in enumerate(numpy.ix_(*arrays)):
        arr[..., i] = array
    return numpy.hsplit(arr.reshape(-1, array_count), array_count)


def _cross_join(left, right):
    """
    A cross join is the cartesian product of two tables - this usually isn't very
    useful, but it does allow you to the theta joins (non-equi joins)
    """

    def _chunker(seq_1, seq_2, size):
        """
        Chunk two equal length interables into size sized chunks

        This returns a generator.
        """
        return (
            (seq_1[pos : pos + size], seq_2[pos : pos + size]) for pos in range(0, len(seq_1), size)
        )

    from opteryx.third_party.pyarrow_ops import align_tables

    right_columns = Columns(right)
    left_columns = None

    for left_morsel in left.execute():
        if left_columns is None:
            left_columns = Columns(left_morsel)
            new_columns = left_columns + right_columns

        # we break this into small chunks, each cycle will have 500 * rows in the right table
        for left_block in left_morsel.to_batches(max_chunksize=INTERNAL_BATCH_SIZE):
            # blocks don't have column_names, so we need to wrap in a table
            left_block = pyarrow.Table.from_batches([left_block], schema=left_morsel.schema)

            # build two lists, 0 to num_rows for each table
            left_array = numpy.arange(left_block.num_rows, dtype=numpy.int64)
            right_array = numpy.arange(right.num_rows, dtype=numpy.int64)

            # build the cartesian product of the two lists
            left_align, right_align = _cartesian_product(left_array, right_array)

            # CROSS JOINs can create huge tables quickly, this is used to limit the
            # number of records we hold in memory at any time
            for left_chunk, right_chunk in _chunker(left_align, right_align, MAX_JOIN_SIZE):
                # now build the resultant table
                table = align_tables(left_block, right, left_chunk.flatten(), right_chunk.flatten())
                yield new_columns.apply(table)


def _cross_join_unnest(
    left: pyarrow.Table, column, alias: str = None
) -> typing.Generator[pyarrow.Table, None, None]:
    """
    Cross join on an unnested column. This function reads a row, creates a new column by unnesting
    a column in that row, and then repeats the row for each unnested value. It then returns the
    resulting table.

    Args:
        left: An iterator of `pyarrow.Table` objects to cross join on.
        column: The column to unnest.
        alias: An optional string to use as the column name for the unnested column.

    Returns:
        A generator that yields the resulting `pyarrow.Table` objects.
    """

    if column.token_type != NodeType.IDENTIFIER:
        raise ValueError("The `column` argument must be a valid identifier.")

    if column.token_type != NodeType.IDENTIFIER:
        raise NotImplementedError("Can only CROSS JOIN UNNEST on a field")

    metadata = None
    column_type = None

    if alias is None:
        alias = f"UNNEST({column.value})"

    for left_morsel in left.execute():
        if metadata is None:
            metadata = Columns(left_morsel)
            metadata.add_column(alias)
            unnest_column = metadata.get_column_from_alias(column.value, only_one=True)

        # we break this into small chunks otherwise we very quickly run into memory issues
        for left_block in left_morsel.to_batches(max_chunksize=INTERNAL_BATCH_SIZE):
            # Get the column we're going to UNNEST
            column_data = left_block[unnest_column]
            if column_type is None:
                column_type = column_data.type.value_type

            # Create a list of indexes, this will look something like this:
            # [1,1,1,2,2,2,3,3,3]
            # Where the number of times a number is repeated, is the length of the list
            # we're going to UNNEST for that row
            indexes = []
            new_column = []
            for i, value in enumerate(column_data):
                # if the value isn't valid, we can't UNNEST it
                if value.is_valid and len(value.values) != 0:
                    indexes.extend([i] * len(value))
                    new_column.extend(value)
                else:
                    indexes.append(i)
                    new_column.append(None)

            if len(indexes) == 0:
                continue

            # Strings need special treatment to avoid them being coerced into a list
            # of characters
            new_column = [(v.as_py() if hasattr(v, "as_py") else v for v in new_column)]

            # Using the indexes above, repeat the rows of the source data
            new_block = left_block.take(indexes)
            # We can't append columns to batches, so we need to convert to a table
            new_block = pyarrow.Table.from_batches([new_block])
            # Append the column we created above, to the table with the repeated rows
            new_block = pyarrow.Table.append_column(new_block, alias, new_column)

            # if the entire columns is nulls, the schema won't match
            column_index = new_block.column_names.index(alias)
            schema = new_block.schema
            column = schema.field(column_index)
            if column.type != column_type:
                schema = schema.set(column_index, pyarrow.field(alias, column_type))
                new_block = new_block.cast(target_schema=schema)

            new_block = metadata.apply(new_block)
            yield new_block


class CrossJoinNode(BasePlanNode):
    """
    Implements a SQL CROSS JOIN and CROSS JOIN UNNEST
    """

    def __init__(self, properties: QueryProperties, **config):
        super().__init__(properties=properties)
        self._right_table = config.get("right_table")
        self._join_type = config.get("join_type", "CrossJoin")

    @property
    def name(self):  # pragma: no cover
        return "Cross Join"

    @property
    def config(self):  # pragma: no cover
        if self._join_type == "CrossJoinUnnest":
            return "UNNEST()"
        return ""

    def execute(self) -> typing.Iterable:
        if len(self._producers) != 2:  # pragma: no cover
            raise SqlError(f"{self.name} expects two producers")

        left_node = self._producers[0]  # type:ignore
        right_node = self._producers[1]  # type:ignore

        if self._join_type == "CrossJoin":
            self._right_table = pyarrow.concat_tables(
                right_node.execute(), promote=True
            )  # type:ignore

            yield from _cross_join(left_node, self._right_table)

        elif self._join_type == "CrossJoinUnnest":
            function = right_node.dataset["function"]
            args = right_node.dataset["args"]

            if function != "unnest":
                raise SqlError(f"I was expecting 'UNNEST' but I got `{function}`")

            yield from _cross_join_unnest(left=left_node, column=args[0], alias=right_node.alias)



"""
This is a initial impementation of a query optimizer.

It is implemented as a set of rules which are executed in turn. It is a heuristic,
no regrets optimizer and rewriter.
"""
import time

from opteryx.components.tree_rewriter import rules
from opteryx.shared.query_statistics import QueryStatistics

RULESET: list = [
    # left joins
    rules.move_literal_join_filters,
    rules.apply_demorgans_law,
    rules.eliminate_negations,
    rules.split_conjunctive_predicates,  # run after eliminate_negations
    rules.eliminate_fixed_function_evaluations,  # run before constant evaluations
    rules.eliminate_constant_evaluations,
    rules.predicate_pushdown,
    rules.defragment_morsels,
    rules.use_heap_sort,
]


def tree_rewriter(plan, properties):
    stats = QueryStatistics(properties.qid)
    start = time.monotonic_ns()

    for rule in RULESET:
        plan = rule(plan, properties)

    stats.time_rewriting += time.monotonic_ns() - start

    return plan



import re


def remove_comments(string):  # pragma: no cover
    """
    Remove comments from the string
    """
    # first group captures quoted strings (double, single or back tick)
    # second group captures comments (//single-line or /* multi-line */)
    pattern = r"(\"[^\"]\"|\'[^\']\'|\`[^\`]\`)|(/\*[^\*/]*\*/|--[^\r\n]*$)"
    regex = re.compile(pattern, re.MULTILINE | re.DOTALL)

    def _replacer(match):
        # if the 2nd group (capturing comments) is not None,
        # it means we have captured a non-quoted (real) comment string.
        if match.group(2) is not None:
            return ""  # so we will return empty to remove the comment
        # otherwise, we will return the 1st group
        return match.group(1)  # captured quoted-string

    return regex.sub(_replacer, string)


def clean_statement(string):  # pragma: no cover
    """
    Remove carriage returns and all whitespace to single spaces.

    Avoid removing whitespace in quoted strings.
    """
    pattern = r"(\"[^\"]\"|\'[^\']\'|\`[^\`]\`)|(\r\n\t\f\v+)"
    regex = re.compile(pattern, re.MULTILINE | re.DOTALL)

    def _replacer(match):
        if match.group(2) is not None:
            return " "
        return match.group(1)  # captured quoted-string

    return regex.sub(_replacer, string).strip()
